import numpy as np
from numpy.testing import assert_almost_equal
from menpo.math import eigenvalue_decomposition, pca, ipca

# Positive semi-definite matrix
cov_matrix = np.array([[3, 1], [1, 3]])
# Data values taken from:
# http://www.cs.otago.ac.nz/cosc453/student_tutorials/principal_components.pdf
# Tested values are equal
large_samples_data_matrix = np.array([[2.5, 2.4],
                                      [0.5, 0.7],
                                      [2.2, 2.9],
                                      [1.9, 2.2],
                                      [3.1, 3.0],
                                      [2.3, 2.7],
                                      [2.0, 1.6],
                                      [1.0, 1.1],
                                      [1.5, 1.6],
                                      [1.1, 0.9]])
centered_eigenvectors_s = np.array([[0.6778734, 0.73517866],
                                    [-0.73517866, 0.6778734]])
non_centered_eigenvectors_s = np.array([[0.68647784, 0.72715072],
                                        [-0.72715072, 0.68647784]])
mean_vector_s = np.array([1.81, 1.91])
eigenvalues_no_centre_s = np.array([8.97738481, 0.04928186])
eigenvalues_centered_s = np.array([1.28402771, 0.0490834])

centered_eigenvectors_f = np.array([[-0.09901475, 0.19802951, 0.69310328,
                                     0.29704426, -0.09901475, 0.39605902,
                                     -0.39605902, 0.09901475, 0.09901475,
                                     -0.19802951]])
non_centered_eigenvectors_f = np.array(
    [[0.38507927, 0.09481302, 0.40261598, 0.32307722, 0.4794398, 0.39407387,
      0.28217662, 0.16535718, 0.24399096, 0.15681507],
     [-0.25575629, 0.17561812, 0.58718113, 0.19319469, -0.29239933, 0.27184299,
      -0.5344514, 0.04200527, 0.01146941, -0.27333287]])
mean_vector_f = np.array([2.45, 0.6, 2.55, 2.05, 3.05,
                          2.5, 1.8, 1.05, 1.55, 1.])
eigenvalues_no_centre_f = np.array([80.79646326, 0.44353674])
eigenvalues_centered_f = np.array([0.51])


def test_pcd_samples_nocentre():
    output = pca(large_samples_data_matrix, centre=False)
    eigenvectors, eigenvalues, mean_vector = output

    assert_almost_equal(eigenvalues, eigenvalues_no_centre_s)
    assert_almost_equal(eigenvectors, non_centered_eigenvectors_s)
    assert_almost_equal(mean_vector, [0.0, 0.0])


def test_pcd_samples_yescentre():
    output = pca(large_samples_data_matrix, centre=True)
    eigenvectors, eigenvalues, mean_vector = output

    assert_almost_equal(eigenvalues, eigenvalues_centered_s)
    assert_almost_equal(eigenvectors, centered_eigenvectors_s)
    assert_almost_equal(mean_vector, mean_vector_s)


def test_pcd_features_nocentre():
    output = pca(large_samples_data_matrix.T, centre=False)
    eigenvectors, eigenvalues, mean_vector = output

    assert_almost_equal(eigenvalues, eigenvalues_no_centre_f)
    assert_almost_equal(eigenvectors, non_centered_eigenvectors_f)
    assert_almost_equal(mean_vector, np.zeros(10))


def test_pcd_features_nocentre_inplace():
    # important to copy as this will now destructively effect the input data
    # matrix (due to inplace)
    output = pca(large_samples_data_matrix.T.copy(), centre=False,
                 inplace=True)
    eigenvectors, eigenvalues, mean_vector = output

    assert_almost_equal(eigenvalues, eigenvalues_no_centre_f)
    assert_almost_equal(eigenvectors, non_centered_eigenvectors_f)
    assert_almost_equal(mean_vector, np.zeros(10))


def test_pcd_features_yescentre():
    output = pca(large_samples_data_matrix.T, centre=True)
    eigenvectors, eigenvalues, mean_vector = output

    assert_almost_equal(eigenvalues, eigenvalues_centered_f)
    assert_almost_equal(eigenvectors, centered_eigenvectors_f)
    assert_almost_equal(mean_vector, mean_vector_f)


def test_eigenvalue_decomposition_default_epsilon():
    pos_eigenvectors, pos_eigenvalues = eigenvalue_decomposition(cov_matrix)

    assert_almost_equal(pos_eigenvalues, [4.0, 2.0])
    sqrt_one_over_2 = np.sqrt(2.0) / 2.0
    assert_almost_equal(pos_eigenvectors, [[sqrt_one_over_2, -sqrt_one_over_2],
                                           [sqrt_one_over_2, sqrt_one_over_2]])


def test_eigenvalue_decomposition_large_epsilon():
    pos_eigenvectors, pos_eigenvalues = eigenvalue_decomposition(cov_matrix,
                                                                 eps=0.5)

    assert_almost_equal(pos_eigenvalues, [4.0])
    sqrt_one_over_2 = np.sqrt(2.0) / 2.0
    assert_almost_equal(pos_eigenvectors,
                        [[sqrt_one_over_2], [sqrt_one_over_2]])


def test_ipca_samples_yescentre():
    n_a = large_samples_data_matrix.shape[0] // 2
    A = large_samples_data_matrix[:n_a, :]
    U_a, l_a, m_a = pca(A, centre=True)

    B = large_samples_data_matrix[n_a:, :]
    i_U, i_l, i_m = ipca(B, U_a, l_a, n_a, m_a=m_a)

    b_U, b_l, b_m = pca(large_samples_data_matrix, centre=True)

    assert_almost_equal(np.abs(i_U), np.abs(b_U))
    assert_almost_equal(i_l, b_l)
    assert_almost_equal(i_m, b_m)


def test_ipca_samples_nocentre():
    n_a = large_samples_data_matrix.shape[0] // 2
    A = large_samples_data_matrix[:n_a, :]
    U_a, l_a, m_a = pca(A, centre=False)

    B = large_samples_data_matrix[n_a:, :]
    i_U, i_l, i_m = ipca(B, U_a, l_a, n_a, m_a=m_a)

    b_U, b_l, b_m = pca(large_samples_data_matrix, centre=False)

    assert_almost_equal(np.abs(i_U), np.abs(b_U))
    assert_almost_equal(i_l, b_l)
    assert_almost_equal(i_m, b_m)


def test_ipca_features_yescentre():
    C = np.vstack((large_samples_data_matrix.T, large_samples_data_matrix.T))

    n_a = C.shape[0] // 2
    A = C[:n_a, :]
    U_a, l_a, m_a = pca(A, centre=True)

    B = C[n_a:, :]
    i_U, i_l, i_m = ipca(B, U_a, l_a, n_a, m_a=m_a)

    b_U, b_l, b_m = pca(C, centre=True)

    assert_almost_equal(np.abs(i_U), np.abs(b_U))
    assert_almost_equal(i_l, b_l)
    assert_almost_equal(i_m, b_m)


def test_ipca_features_nocentre():
    C = np.vstack((large_samples_data_matrix.T, large_samples_data_matrix.T))

    n_a = C.shape[0] // 2
    A = C[:n_a, :]
    U_a, l_a, m_a = pca(A, centre=False)

    B = C[n_a:, :]
    i_U, i_l, i_m = ipca(B, U_a, l_a, n_a, m_a=m_a)

    b_U, b_l, b_m = pca(C, centre=False)

    assert_almost_equal(np.abs(i_U), np.abs(b_U))
    assert_almost_equal(i_l, b_l)
    assert_almost_equal(i_m, b_m)



try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
import numpy as np

from menpo.base import MenpoMissingDependencyError


class Menpo3dMissingError(MenpoMissingDependencyError):
    r"""
    Exception that is thrown when an attempt is made to import a 3D
    visualisation method, but 'menpo3d' is not installed.
    """

    def __init__(self, actual_missing_import_name):
        super(Menpo3dMissingError, self).__init__(actual_missing_import_name)
        self.message += ("\nThis import is required in order to use the "
                         "'menpo3d' package")


class MenpowidgetsMissingError(MenpoMissingDependencyError):
    r"""
    Exception that is thrown when an attempt is made to import a widget, but
    'menpowidgets' is not installed.
    """

    def __init__(self, actual_missing_import_name):
        super(MenpowidgetsMissingError, self).__init__(actual_missing_import_name)
        self.message += ("\nThis import is required in order to use the "
                         "'menpowidgets' package")


class Renderer(object):
    r"""
    Abstract class for rendering visualizations. Framework specific
    implementations of these classes are made in order to separate
    implementation cleanly from the rest of the code.

    It is assumed that the renderers follow some form of stateful pattern for
    rendering to Figures. Therefore, the major interface for rendering involves
    providing a `figure_id` or a `bool` about whether a new figure should be
    used. If neither are provided then the default state of the rendering engine
    is assumed to be maintained.

    Providing both a ``figure_id`` and ``new_figure == True`` is not a valid
    state.

    Parameters
    ----------
    figure_id : `object`
        A figure id. Could be any valid object that identifies a figure in a
        given framework (`str`, `int`, `float`, etc.).
    new_figure : `bool`
        Whether the rendering engine should create a new figure.

    Raises
    ------
    ValueError
        It is not valid to provide a figure id AND request a new figure to
        be rendered on.
    """

    def __init__(self, figure_id, new_figure):
        if figure_id is not None and new_figure:
            raise ValueError("Conflicting arguments. figure_id cannot be "
                             "specified if the new_figure flag is True")

        self.figure_id = figure_id
        self.new_figure = new_figure
        self.figure = self.get_figure()

    def render(self, **kwargs):
        r"""
        Abstract method to be overridden by the renderer. This will implement
        the actual rendering code for a given object class.

        Parameters
        ----------
        kwargs : `dict`
            Passed through to specific rendering engine.

        Returns
        -------
        viewer : :map:`Renderer`
            Pointer to `self`.
        """
        pass

    def get_figure(self):
        r"""
        Abstract method for getting the correct figure to render on. Should
        also set the correct `figure_id` for the figure.

        Returns
        -------
        figure : `object`
            The figure object that the renderer will render on.
        """
        pass

    def save_figure(self, **kwargs):
        r"""
        Abstract method for saving the figure of the current `figure_id` to
        file. It will implement the actual saving code for a given object class.

        Parameters
        ----------
        kwargs : `dict`
            Options to be set when saving the figure to file.
        """
        pass

    def clear_figure(self):
        r"""
        Abstract method for clearing the current figure.
        """
        pass

    def force_draw(self):
        r"""
        Abstract method for forcing the current figure to render.
        """
        pass


class viewwrapper(object):
    r"""
    This class abuses the Python descriptor protocol in order to dynamically
    change the view method at runtime. Although this is more obviously achieved
    through inheritance, the view methods practically amount to syntactic sugar
    and so we want to maintain a single view method per class. We do not want
    to add the mental overhead of implementing different 2D and 3D PointCloud
    classes for example, since, outside of viewing, their implementations would
    be identical.

    Also note that we could have separated out viewing entirely and made the
    check there, but the view method is an important paradigm in menpo that
    we want to maintain.

    Therefore, this function cleverly (and obscurely) returns the correct
    view method for the dimensionality of the given object.
    """

    def __init__(self, wrapped_func):
        fname = wrapped_func.__name__
        self._2d_fname = '_{}_2d'.format(fname)
        self._3d_fname = '_{}_3d'.format(fname)

    def __get__(self, instance, instancetype):
        if instance.n_dims == 2:
            return getattr(instance, self._2d_fname)
        elif instance.n_dims == 3:
            return getattr(instance, self._3d_fname)
        else:
            def raise_not_supported(*args, **kwargs):
                r"""
                Viewing of objects with greater than 3 dimensions is not
                currently possible.
                """
                raise ValueError('Viewing of objects with greater than 3 '
                                 'dimensions is not currently possible.')

            return raise_not_supported


class Viewable(object):
    r"""
    Abstract interface for objects that can visualize themselves. This assumes
    that the class has dimensionality as the view method checks the ``n_dims``
    property to wire up the correct view method.
    """

    @viewwrapper
    def view(self):
        r"""
        Abstract method for viewing. See the :map:`viewwrapper` documentation
        for an explanation of how the `view` method works.
        """
        pass

    def _view_2d(self, **kwargs):
        raise NotImplementedError('2D Viewing is not supported.')

    def _view_3d(self, **kwargs):
        raise NotImplementedError('3D Viewing is not supported.')


class LandmarkableViewable(object):
    r"""
    Mixin for :map:`Landmarkable` and :map:`Viewable` objects. Provides a
    single helper method for viewing Landmarks and `self` on the same figure.
    """

    @viewwrapper
    def view_landmarks(self, **kwargs):
        pass

    def _view_landmarks_2d(self, **kwargs):
        raise NotImplementedError('2D Landmark Viewing is not supported.')

    def _view_landmarks_3d(self, **kwargs):
        raise NotImplementedError('3D Landmark Viewing is not supported.')


from menpo.visualize.viewmatplotlib import (
    MatplotlibImageViewer2d, MatplotlibImageSubplotsViewer2d,
    MatplotlibLandmarkViewer2d, MatplotlibAlignmentViewer2d,
    MatplotlibGraphPlotter, MatplotlibMultiImageViewer2d,
    MatplotlibMultiImageSubplotsViewer2d, MatplotlibPointGraphViewer2d)

# Default importer types
PointGraphViewer2d = MatplotlibPointGraphViewer2d
LandmarkViewer2d = MatplotlibLandmarkViewer2d
ImageViewer2d = MatplotlibImageViewer2d
ImageSubplotsViewer2d = MatplotlibImageSubplotsViewer2d

AlignmentViewer2d = MatplotlibAlignmentViewer2d
GraphPlotter = MatplotlibGraphPlotter
MultiImageViewer2d = MatplotlibMultiImageViewer2d
MultiImageSubplotsViewer2d = MatplotlibMultiImageSubplotsViewer2d


class ImageViewer(object):
    r"""
    Base :map:`Image` viewer that abstracts away dimensionality. It can
    visualize multiple channels of an image in subplots.

    Parameters
    ----------
    figure_id : `object`
        A figure id. Could be any valid object that identifies a figure in a
        given framework (`str`, `int`, `float`, etc.).
    new_figure : `bool`
        Whether the rendering engine should create a new figure.
    dimensions : {``2``, ``3``} `int`
        The number of dimensions in the image.
    pixels : ``(N, D)`` `ndarray`
        The pixels to render.
    channels: `int` or `list` or ``'all'`` or `None`
        A specific selection of channels to render. The user can choose either
        a single or multiple channels. If ``'all'``, render all channels in
        subplot mode. If `None` and image is not greyscale or RGB, render all
        channels in subplots. If `None` and image is greyscale or RGB, then do
        not plot channels in different subplots.
    mask: ``(N, D)`` `ndarray`
        A `bool` mask to be applied to the image. All points outside the
        mask are set to ``0``.
    """

    def __init__(self, figure_id, new_figure, dimensions, pixels,
                 channels=None, mask=None):
        if len(pixels.shape) == 3 and pixels.shape[0] == 3:
            # then probably an RGB image, so ensure the clipped pixels.
            from menpo.image import Image
            image = Image(pixels, copy=False)
            image_clipped = image.clip_pixels()
            pixels = image_clipped.pixels
        else:
            pixels = pixels.copy()
        self.figure_id = figure_id
        self.new_figure = new_figure
        self.dimensions = dimensions
        pixels, self.use_subplots = \
            self._parse_channels(channels, pixels)
        self.pixels = self._masked_pixels(pixels, mask)

        self._flip_image_channels()

    def _flip_image_channels(self):
        if self.pixels.ndim == 3:
            from menpo.image.base import channels_to_back
            self.pixels = channels_to_back(self.pixels)

    def _parse_channels(self, channels, pixels):
        r"""
        Parse `channels` parameter. If `channels` is `int` or `list`, keep it as
        is. If `channels` is ``'all'``, return a `list` of all the image's
        channels. If `channels` is `None`, return the minimum between an
        `upper_limit` and the image's number of channels. If image is greyscale
        or RGB and `channels` is `None`, then do not plot channels in different
        subplots.

        Parameters
        ----------
        channels : `int` or `list` or ``'all'`` or `None`
            A specific selection of channels to render.
        pixels : ``(N, D)`` `ndarray`
            The image's pixels to render.

        Returns
        -------
        pixels : ``(N, D)`` `ndarray`
            The pixels to be visualized.
        use_subplots : `bool`
            Whether to visualize using subplots.
        """
        # Flag to trigger ImageSubplotsViewer2d or ImageViewer2d
        use_subplots = True
        n_channels = pixels.shape[0]
        if channels is None:
            if n_channels == 1:
                pixels = pixels[0, ...]
                use_subplots = False
            elif n_channels == 3:
                use_subplots = False
        elif channels != 'all':
            if isinstance(channels, Iterable):
                if len(channels) == 1:
                    pixels = pixels[channels[0], ...]
                    use_subplots = False
                else:
                    pixels = pixels[channels, ...]
            else:
                pixels = pixels[channels, ...]
                use_subplots = False

        return pixels, use_subplots

    def _masked_pixels(self, pixels, mask):
        r"""
        Return the masked pixels using a given `bool` mask. In order to make
        sure that the non-masked pixels are visualized in white, their value
        is set to the maximum of pixels.

        Parameters
        ----------
        pixels : ``(N, D)`` `ndarray`
            The image's pixels to render.
        mask: ``(N, D)`` `ndarray`
            A `bool` mask to be applied to the image. All points outside the
            mask are set to the image max. If mask is `None`, then the initial
            pixels are returned.

        Returns
        -------
        masked_pixels : ``(N, D)`` `ndarray`
            The masked pixels.
        """
        if mask is not None:
            nanmax = np.nanmax(pixels)
            pixels[..., ~mask] = nanmax + (0.01 * nanmax)
        return pixels

    def render(self, **kwargs):
        r"""
        Select the correct type of image viewer for the given image
        dimensionality.

        Parameters
        ----------
        kwargs : `dict`
            Passed through to image viewer.

        Returns
        -------
        viewer : :map:`Renderer`
            The rendering object.

        Raises
        ------
        ValueError
            Only 2D images are supported.
        """
        if self.dimensions == 2:
            if self.use_subplots:
                return ImageSubplotsViewer2d(self.figure_id, self.new_figure,
                                             self.pixels).render(**kwargs)
            else:
                return ImageViewer2d(self.figure_id, self.new_figure,
                                     self.pixels).render(**kwargs)
        else:
            raise ValueError("Only 2D images are currently supported")


def view_image_landmarks(image, channels, masked, group,
                         with_labels, without_labels, figure_id, new_figure,
                         interpolation, cmap_name, alpha, render_lines,
                         line_colour, line_style, line_width,
                         render_markers, marker_style, marker_size,
                         marker_face_colour, marker_edge_colour,
                         marker_edge_width, render_numbering,
                         numbers_horizontal_align, numbers_vertical_align,
                         numbers_font_name, numbers_font_size,
                         numbers_font_style, numbers_font_weight,
                         numbers_font_colour, render_legend, legend_title,
                         legend_font_name, legend_font_style, legend_font_size,
                         legend_font_weight, legend_marker_scale,
                         legend_location, legend_bbox_to_anchor,
                         legend_border_axes_pad, legend_n_columns,
                         legend_horizontal_spacing, legend_vertical_spacing,
                         legend_border, legend_border_padding, legend_shadow,
                         legend_rounded_corners, render_axes, axes_font_name,
                         axes_font_size, axes_font_style, axes_font_weight,
                         axes_x_limits, axes_y_limits, axes_x_ticks,
                         axes_y_ticks, figure_size):
    r"""
    This is a helper method that abstracts away the fact that viewing
    images and masked images is identical apart from the mask. Therefore,
    we do the class check in this method and then proceed identically whether
    the image is masked or not.

    See the documentation for _view_2d on Image or _view_2d on MaskedImage
    for information about the parameters.
    """
    import matplotlib.pyplot as plt

    if not image.has_landmarks:
        raise ValueError('Image does not have landmarks attached, unable '
                         'to view landmarks.')

    # Parse axes limits
    image_axes_x_limits = None
    landmarks_axes_x_limits = axes_x_limits
    if axes_x_limits is None:
        image_axes_x_limits = landmarks_axes_x_limits = [0, image.width - 1]
    image_axes_y_limits = None
    landmarks_axes_y_limits = axes_y_limits
    if axes_y_limits is None:
        image_axes_y_limits = landmarks_axes_y_limits = [0, image.height - 1]

    # Render image
    from menpo.image import MaskedImage
    if isinstance(image, MaskedImage):
        self_view = image.view(figure_id=figure_id, new_figure=new_figure,
                               channels=channels, masked=masked,
                               interpolation=interpolation, cmap_name=cmap_name,
                               alpha=alpha, render_axes=render_axes,
                               axes_x_limits=image_axes_x_limits,
                               axes_y_limits=image_axes_y_limits)
    else:
        self_view = image.view(figure_id=figure_id, new_figure=new_figure,
                               channels=channels, interpolation=interpolation,
                               cmap_name=cmap_name, alpha=alpha,
                               render_axes=render_axes,
                               axes_x_limits=image_axes_x_limits,
                               axes_y_limits=image_axes_y_limits)

    # Render landmarks
    # correct group label in legend
    if group is None and image.landmarks.n_groups == 1:
        group = image.landmarks.group_labels[0]
    landmark_view = None  # initialize viewer object
    # useful in order to visualize the legend only for the last axis object
    render_legend_tmp = False
    for i, ax in enumerate(self_view.axes_list):
        # set current axis
        plt.sca(ax)
        # show legend only for the last axis object
        if i == len(self_view.axes_list) - 1:
            render_legend_tmp = render_legend

        # viewer
        landmark_view = image.landmarks[group].view(
            with_labels=with_labels, without_labels=without_labels,
            group=group, figure_id=self_view.figure_id, new_figure=False,
            image_view=True, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour,
            render_legend=render_legend_tmp, legend_title=legend_title,
            legend_font_name=legend_font_name,
            legend_font_style=legend_font_style,
            legend_font_size=legend_font_size,
            legend_font_weight=legend_font_weight,
            legend_marker_scale=legend_marker_scale,
            legend_location=legend_location,
            legend_bbox_to_anchor=legend_bbox_to_anchor,
            legend_border_axes_pad=legend_border_axes_pad,
            legend_n_columns=legend_n_columns,
            legend_horizontal_spacing=legend_horizontal_spacing,
            legend_vertical_spacing=legend_vertical_spacing,
            legend_border=legend_border,
            legend_border_padding=legend_border_padding,
            legend_shadow=legend_shadow,
            legend_rounded_corners=legend_rounded_corners,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight,
            axes_x_limits=landmarks_axes_x_limits,
            axes_y_limits=landmarks_axes_y_limits, axes_x_ticks=axes_x_ticks,
            axes_y_ticks=axes_y_ticks, figure_size=figure_size)

    return landmark_view


class MultipleImageViewer(ImageViewer):

    def __init__(self, figure_id, new_figure, dimensions, pixels_list,
                 channels=None, mask=None):
        super(MultipleImageViewer, self).__init__(
            figure_id, new_figure, dimensions, pixels_list[0],
            channels=channels, mask=mask)
        pixels_list = [self._parse_channels(channels, p)[0]
                       for p in pixels_list]
        self.pixels_list = [self._masked_pixels(p, mask)
                            for p in pixels_list]

    def render(self, **kwargs):
        if self.dimensions == 2:
            if self.use_subplots:
                MultiImageSubplotsViewer2d(self.figure_id, self.new_figure,
                                           self.pixels_list).render(**kwargs)
            else:
                return MultiImageViewer2d(self.figure_id, self.new_figure,
                                          self.pixels_list).render(**kwargs)
        else:
            raise ValueError("Only 2D images are currently supported")


def plot_curve(x_axis, y_axis, figure_id=None, new_figure=True,
               legend_entries=None, title='', x_label='', y_label='',
               axes_x_limits=0., axes_y_limits=None, axes_x_ticks=None,
               axes_y_ticks=None, render_lines=True, line_colour=None,
               line_style='-', line_width=1, render_markers=True,
               marker_style='o', marker_size=5, marker_face_colour=None,
               marker_edge_colour='k', marker_edge_width=1., render_legend=True,
               legend_title='', legend_font_name='sans-serif',
               legend_font_style='normal', legend_font_size=10,
               legend_font_weight='normal', legend_marker_scale=None,
               legend_location=2, legend_bbox_to_anchor=(1.05, 1.),
               legend_border_axes_pad=None, legend_n_columns=1,
               legend_horizontal_spacing=None, legend_vertical_spacing=None,
               legend_border=True, legend_border_padding=None,
               legend_shadow=False, legend_rounded_corners=False,
               render_axes=True, axes_font_name='sans-serif', axes_font_size=10,
               axes_font_style='normal', axes_font_weight='normal',
               figure_size=(7, 7), render_grid=True, grid_line_style='--',
               grid_line_width=1):
    r"""
    Plot a single or multiple curves on the same figure.

    Parameters
    ----------
    x_axis : `list` or `array`
        The values of the horizontal axis. They are common for all curves.
    y_axis : `list` of `lists` or `arrays`
        A `list` with `lists` or `arrays` with the values of the vertical axis
        for each curve.
    figure_id : `object`, optional
        The id of the figure to be used.
    new_figure : `bool`, optional
        If ``True``, a new figure is created.
    legend_entries : `list of `str` or ``None``, optional
        If `list` of `str`, it must have the same length as `errors` `list` and
        each `str` will be used to name each curve. If ``None``, the CED curves
        will be named as `'Curve %d'`.
    title : `str`, optional
        The figure's title.
    x_label : `str`, optional
        The label of the horizontal axis.
    y_label : `str`, optional
        The label of the vertical axis.
    axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
        The limits of the x axis. If `float`, then it sets padding on the
        right and left of the graph as a percentage of the curves' width. If
        `tuple` or `list`, then it defines the axis limits. If ``None``, then the
        limits are set automatically.
    axes_y_limits : `float` or (`float`, `float`) or ``None``, optional
        The limits of the y axis. If `float`, then it sets padding on the
        top and bottom of the graph as a percentage of the curves' height. If
        `tuple` or `list`, then it defines the axis limits. If ``None``, then the
        limits are set automatically.
    axes_x_ticks : `list` or `tuple` or ``None``, optional
        The ticks of the x axis.
    axes_y_ticks : `list` or `tuple` or ``None``, optional
        The ticks of the y axis.
    render_lines : `bool` or `list` of `bool`, optional
        If ``True``, the line will be rendered. If `bool`, this value will be
        used for all curves. If `list`, a value must be specified for each
        curve, thus it must have the same length as `y_axis`.
    line_colour : `colour` or `list` of `colour` or ``None``, optional
        The colour of the lines. If not a `list`, this value will be
        used for all curves. If `list`, a value must be specified for each
        curve, thus it must have the same length as `y_axis`. If ``None``, the
        colours will be linearly sampled from jet colormap.
        Example `colour` options are ::

                {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                or
                (3, ) ndarray

    line_style : ``{'-', '--', '-.', ':'}`` or `list` of those, optional
        The style of the lines. If not a `list`, this value will be used for all
        curves. If `list`, a value must be specified for each curve, thus it must
        have the same length as `y_axis`.
    line_width : `float` or `list` of `float`, optional
        The width of the lines. If `float`, this value will be used for all
        curves. If `list`, a value must be specified for each curve, thus it must
        have the same length as `y_axis`.
    render_markers : `bool` or `list` of `bool`, optional
        If ``True``, the markers will be rendered. If `bool`, this value will be
        used for all curves. If `list`, a value must be specified for each
        curve, thus it must have the same length as `y_axis`.
    marker_style : `marker` or `list` of `markers`, optional
        The style of the markers. If not a `list`, this value will be used for
        all curves. If `list`, a value must be specified for each curve, thus it
        must have the same length as `y_axis`.
        Example `marker` options ::

                {'.', ',', 'o', 'v', '^', '<', '>', '+', 'x', 'D', 'd', 's',
                 'p', '*', 'h', 'H', '1', '2', '3', '4', '8'}

    marker_size : `int` or `list` of `int`, optional
        The size of the markers in points. If `int`, this value will be used
        for all curves. If `list`, a value must be specified for each curve, thus
        it must have the same length as `y_axis`.
    marker_face_colour : `colour` or `list` of `colour` or ``None``, optional
        The face (filling) colour of the markers. If not a `list`, this value
        will be used for all curves. If `list`, a value must be specified for
        each curve, thus it must have the same length as `y_axis`. If ``None``,
        the colours will be linearly sampled from jet colormap.
        Example `colour` options are ::

                {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                or
                (3, ) ndarray

    marker_edge_colour : `colour` or `list` of `colour` or ``None``, optional
        The edge colour of the markers. If not a `list`, this value will be used
        for all curves. If `list`, a value must be specified for each curve, thus
        it must have the same length as `y_axis`. If ``None``, the colours will
        be linearly sampled from jet colormap.
        Example `colour` options are ::

                {'r', 'g', 'b', 'c', 'm', 'k', 'w'}
                or
                (3, ) ndarray

    marker_edge_width : `float` or `list` of `float`, optional
        The width of the markers' edge. If `float`, this value will be used for
        all curves. If `list`, a value must be specified for each curve, thus it
        must have the same length as `y_axis`.
    render_legend : `bool`, optional
        If ``True``, the legend will be rendered.
    legend_title : `str`, optional
        The title of the legend.
    legend_font_name : See below, optional
        The font of the legend.
        Example options ::

            {'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'}
            
    legend_font_style : ``{'normal', 'italic', 'oblique'}``, optional
        The font style of the legend.
    legend_font_size : `int`, optional
        The font size of the legend.
    legend_font_weight : See below, optional
        The font weight of the legend.
        Example options ::

            {'ultralight', 'light', 'normal', 'regular', 'book', 'medium',
             'roman', 'semibold', 'demibold', 'demi', 'bold', 'heavy',
             'extra bold', 'black'}

    legend_marker_scale : `float`, optional
        The relative size of the legend markers with respect to the original
    legend_location : `int`, optional
        The location of the legend. The predefined values are:

        =============== ===
        'best'          0
        'upper right'   1
        'upper left'    2
        'lower left'    3
        'lower right'   4
        'right'         5
        'center left'   6
        'center right'  7
        'lower center'  8
        'upper center'  9
        'center'        10
        =============== ===

    legend_bbox_to_anchor : (`float`, `float`), optional
        The bbox that the legend will be anchored.
    legend_border_axes_pad : `float`, optional
        The pad between the axes and legend border.
    legend_n_columns : `int`, optional
        The number of the legend's columns.
    legend_horizontal_spacing : `float`, optional
        The spacing between the columns.
    legend_vertical_spacing : `float`, optional
        The vertical space between the legend entries.
    legend_border : `bool`, optional
        If ``True``, a frame will be drawn around the legend.
    legend_border_padding : `float`, optional
        The fractional whitespace inside the legend border.
    legend_shadow : `bool`, optional
        If ``True``, a shadow will be drawn behind legend.
    legend_rounded_corners : `bool`, optional
        If ``True``, the frame's corners will be rounded (fancybox).
    render_axes : `bool`, optional
        If ``True``, the axes will be rendered.
    axes_font_name : See below, optional
        The font of the axes.
        Example options ::

            {'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'}

    axes_font_size : `int`, optional
        The font size of the axes.
    axes_font_style : ``{'normal', 'italic', 'oblique'}``, optional
        The font style of the axes.
    axes_font_weight : See below, optional
        The font weight of the axes.
        Example options ::

            {'ultralight', 'light', 'normal', 'regular', 'book', 'medium',
             'roman', 'semibold', 'demibold', 'demi', 'bold', 'heavy',
             'extra bold', 'black'}

    figure_size : (`float`, `float`) or ``None``, optional
        The size of the figure in inches.
    render_grid : `bool`, optional
        If ``True``, the grid will be rendered.
    grid_line_style : ``{'-', '--', '-.', ':'}``, optional
        The style of the grid lines.
    grid_line_width : `float`, optional
        The width of the grid lines.

    Raises
    ------
    ValueError
        legend_entries list has different length than y_axis list

    Returns
    -------
    viewer : :map:`GraphPlotter`
        The viewer object.
    """
    from menpo.visualize import GraphPlotter

    # check y_axis
    if not isinstance(y_axis, list):
        y_axis = [y_axis]

    # check legend_entries
    if legend_entries is not None and len(legend_entries) != len(y_axis):
        raise ValueError('legend_entries list has different length than y_axis '
                         'list')

    # render
    return GraphPlotter(figure_id=figure_id, new_figure=new_figure,
                        x_axis=x_axis, y_axis=y_axis, title=title,
                        legend_entries=legend_entries, x_label=x_label,
                        y_label=y_label, x_axis_limits=axes_x_limits,
                        y_axis_limits=axes_y_limits, x_axis_ticks=axes_x_ticks,
                        y_axis_ticks=axes_y_ticks).render(
        render_lines=render_lines, line_colour=line_colour,
        line_style=line_style, line_width=line_width,
        render_markers=render_markers, marker_style=marker_style,
        marker_size=marker_size, marker_face_colour=marker_face_colour,
        marker_edge_colour=marker_edge_colour,
        marker_edge_width=marker_edge_width, render_legend=render_legend,
        legend_title=legend_title, legend_font_name=legend_font_name,
        legend_font_style=legend_font_style, legend_font_size=legend_font_size,
        legend_font_weight=legend_font_weight,
        legend_marker_scale=legend_marker_scale,
        legend_location=legend_location,
        legend_bbox_to_anchor=legend_bbox_to_anchor,
        legend_border_axes_pad=legend_border_axes_pad,
        legend_n_columns=legend_n_columns,
        legend_horizontal_spacing=legend_horizontal_spacing,
        legend_vertical_spacing=legend_vertical_spacing,
        legend_border=legend_border,
        legend_border_padding=legend_border_padding,
        legend_shadow=legend_shadow,
        legend_rounded_corners=legend_rounded_corners, render_axes=render_axes,
        axes_font_name=axes_font_name, axes_font_size=axes_font_size,
        axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
        figure_size=figure_size, render_grid=render_grid,
        grid_line_style=grid_line_style, grid_line_width=grid_line_width)


def render_rectangles_around_patches(centers, patch_shape, axes=None,
                                     image_view=True, line_colour='r',
                                     line_style='-', line_width=1,
                                     interpolation='none'):
    r"""
    Method that renders rectangles of the specified `patch_shape` centered
    around all the points of the provided `centers`.

    Parameters
    ----------
    centers : :map:`PointCloud`
        The centers around which to draw the rectangles.
    patch_shape : `tuple` or `ndarray`, optional
        The size of the rectangle to render.
    axes : `matplotlib.pyplot.axes` object or ``None``, optional
        The axes object on which to render.
    image_view : `bool`, optional
        If ``True`` the rectangles will be viewed as if they are in the image
        coordinate system.
    line_colour : See Below, optional
        The colour of the lines.
        Example options::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    line_style : ``{-, --, -., :}``, optional
        The style of the lines.
    line_width : `float`, optional
        The width of the lines.
    interpolation : See Below, optional
        In case a patch-based image is already rendered on the specified axes,
        this argument controls how tight the rectangles would be to the patches.
        It needs to have the same value as the one used when rendering the
        patches image, otherwise there is the danger that the rectangles won't
        be exactly on the border of the patches. Example options ::

            {none, nearest, bilinear, bicubic, spline16, spline36,
            hanning, hamming, hermite, kaiser, quadric, catrom, gaussian,
            bessel, mitchell, sinc, lanczos}

    """
    import matplotlib.pyplot as plt
    from matplotlib.patches import Rectangle

    # Dictionary with the line styles
    line_style_dict = {'-': 'solid', '--': 'dashed', '-.': 'dashdot',
                       ':': 'dotted'}

    # Get axes object
    if axes is None:
        axes = plt.gca()

    # Need those in order to compute the lower left corner of the rectangle
    half_patch_shape = [patch_shape[0] / 2,
                        patch_shape[1] / 2]

    # Set the view mode
    if image_view:
        xi = 1
        yi = 0
    else:
        xi = 0
        yi = 1

    # Set correct offsets so that the rectangle is tight to the patch
    if interpolation == 'none':
        off_start = 0.5
        off_end = 0.
    else:
        off_start = 1.
        off_end = 0.5

    # Render rectangles
    for p in range(centers.shape[0]):
        xc = np.intp(centers[p, xi] - half_patch_shape[xi]) - off_start
        yc = np.intp(centers[p, yi] - half_patch_shape[yi]) - off_start
        axes.add_patch(Rectangle((xc, yc),
                                 patch_shape[xi] + off_end,
                                 patch_shape[yi] + off_end,
                                 fill=False, edgecolor=line_colour,
                                 linewidth=line_width,
                                 linestyle=line_style_dict[line_style]))


def view_patches(patches, patch_centers, patches_indices=None,
                 offset_index=None, figure_id=None, new_figure=False,
                 background='white', render_patches=True, channels=None,
                 interpolation='none', cmap_name=None, alpha=1.,
                 render_patches_bboxes=True, bboxes_line_colour='r',
                 bboxes_line_style='-', bboxes_line_width=1,
                 render_centers=True, render_lines=True, line_colour=None,
                 line_style='-', line_width=1, render_markers=True,
                 marker_style='o', marker_size=5, marker_face_colour=None,
                 marker_edge_colour=None, marker_edge_width=1.,
                 render_numbering=False, numbers_horizontal_align='center',
                 numbers_vertical_align='bottom',
                 numbers_font_name='sans-serif', numbers_font_size=10,
                 numbers_font_style='normal', numbers_font_weight='normal',
                 numbers_font_colour='k', render_axes=False,
                 axes_font_name='sans-serif', axes_font_size=10,
                 axes_font_style='normal', axes_font_weight='normal',
                 axes_x_limits=None, axes_y_limits=None, axes_x_ticks=None,
                 axes_y_ticks=None, figure_size=(7, 7)):
    r"""
    Method that renders the provided `patches` on a canvas. The user can
    choose whether to render the patch centers (`render_centers`) as well as
    rectangle boundaries around the patches (`render_patches_bboxes`).

    The patches argument can have any of the two formats that are returned
    from the `extract_patches()` and `extract_patches_around_landmarks()`
    methods of the :map:`Image` class. Specifically it can be:

        1. ``(n_center, n_offset, self.n_channels, patch_shape)`` `ndarray`
        2. `list` of ``n_center * n_offset`` :map:`Image` objects

    Parameters
    ----------
    patches : `ndarray` or `list`
        The values of the patches. It can have any of the two formats that are
        returned from the `extract_patches()` and
        `extract_patches_around_landmarks()` methods. Specifically, it can
        either be an ``(n_center, n_offset, self.n_channels, patch_shape)``
        `ndarray` or a `list` of ``n_center * n_offset`` :map:`Image` objects.
    patch_centers : :map:`PointCloud`
        The centers around which to visualize the patches.
    patches_indices : `int` or `list` of `int` or ``None``, optional
        Defines the patches that will be visualized. If ``None``, then all the
        patches are selected.
    offset_index : `int` or ``None``, optional
        The offset index within the provided `patches` argument, thus the index
        of the second dimension from which to sample. If ``None``, then ``0`` is
        used.
    figure_id : `object`, optional
        The id of the figure to be used.
    new_figure : `bool`, optional
        If ``True``, a new figure is created.
    background : ``{'black', 'white'}``, optional
        If ``'black'``, then the background is set equal to the minimum value
        of `patches`. If ``'white'``, then the background is set equal to the
        maximum value of `patches`.
    render_patches : `bool`, optional
        Flag that determines whether to render the patch values.
    channels : `int` or `list` of `int` or ``all`` or ``None``, optional
        If `int` or `list` of `int`, the specified channel(s) will be
        rendered. If ``all``, all the channels will be rendered in subplots.
        If ``None`` and the image is RGB, it will be rendered in RGB mode.
        If ``None`` and the image is not RGB, it is equivalent to ``all``.
    interpolation : See Below, optional
        The interpolation used to render the image. For example, if
        ``bilinear``, the image will be smooth and if ``nearest``, the
        image will be pixelated. Example options ::

            {none, nearest, bilinear, bicubic, spline16, spline36, hanning,
            hamming, hermite, kaiser, quadric, catrom, gaussian, bessel,
            mitchell, sinc, lanczos}

    cmap_name: `str`, optional,
        If ``None``, single channel and three channel images default
        to greyscale and rgb colormaps respectively.
    alpha : `float`, optional
        The alpha blending value, between 0 (transparent) and 1 (opaque).
    render_patches_bboxes : `bool`, optional
        Flag that determines whether to render the bounding box lines around the
        patches.
    bboxes_line_colour : See Below, optional
        The colour of the lines.
        Example options::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray
    bboxes_line_style : ``{-, --, -., :}``, optional
        The style of the lines.
    bboxes_line_width : `float`, optional
        The width of the lines.
    render_centers : `bool`, optional
        Flag that determines whether to render the patch centers.
    render_lines : `bool`, optional
        If ``True``, the edges will be rendered.
    line_colour : See Below, optional
        The colour of the lines.
        Example options::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    line_style : ``{-, --, -., :}``, optional
        The style of the lines.
    line_width : `float`, optional
        The width of the lines.
    render_markers : `bool`, optional
        If ``True``, the markers will be rendered.
    marker_style : See Below, optional
        The style of the markers. Example options ::

            {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

    marker_size : `int`, optional
        The size of the markers in points.
    marker_face_colour : See Below, optional
        The face (filling) colour of the markers.
        Example options ::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    marker_edge_colour : See Below, optional
        The edge colour of the markers.
        Example options ::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    marker_edge_width : `float`, optional
        The width of the markers' edge.
    render_numbering : `bool`, optional
        If ``True``, the landmarks will be numbered.
    numbers_horizontal_align : ``{center, right, left}``, optional
        The horizontal alignment of the numbers' texts.
    numbers_vertical_align : ``{center, top, bottom, baseline}``, optional
        The vertical alignment of the numbers' texts.
    numbers_font_name : See Below, optional
        The font of the numbers. Example options ::

            {serif, sans-serif, cursive, fantasy, monospace}

    numbers_font_size : `int`, optional
        The font size of the numbers.
    numbers_font_style : ``{normal, italic, oblique}``, optional
        The font style of the numbers.
    numbers_font_weight : See Below, optional
        The font weight of the numbers.
        Example options ::

            {ultralight, light, normal, regular, book, medium, roman,
            semibold, demibold, demi, bold, heavy, extra bold, black}

    numbers_font_colour : See Below, optional
        The font colour of the numbers.
        Example options ::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    render_axes : `bool`, optional
        If ``True``, the axes will be rendered.
    axes_font_name : See Below, optional
        The font of the axes. Example options ::

            {serif, sans-serif, cursive, fantasy, monospace}

    axes_font_size : `int`, optional
        The font size of the axes.
    axes_font_style : ``{normal, italic, oblique}``, optional
        The font style of the axes.
    axes_font_weight : See Below, optional
        The font weight of the axes.
        Example options ::

            {ultralight, light, normal, regular, book, medium, roman,
            semibold,demibold, demi, bold, heavy, extra bold, black}

    axes_x_limits : `float` or (`float`, `float`) or ``None``, optional
        The limits of the x axis. If `float`, then it sets padding on the
        right and left of the shape as a percentage of the shape's width. If
        `tuple` or `list`, then it defines the axis limits. If ``None``, then the
        limits are set automatically.
    axes_y_limits : `float` or (`float`, `float`) or ``None``, optional
        The limits of the y axis. If `float`, then it sets padding on the
        top and bottom of the shape as a percentage of the shape's height. If
        `tuple` or `list`, then it defines the axis limits. If ``None``, then the
        limits are set automatically.
    axes_x_ticks : `list` or `tuple` or ``None``, optional
        The ticks of the x axis.
    axes_y_ticks : `list` or `tuple` or ``None``, optional
        The ticks of the y axis.
    figure_size : (`float`, `float`) `tuple` or ``None`` optional
        The size of the figure in inches.

    Returns
    -------
    viewer : `ImageViewer`
        The image viewing object.
    """
    from menpo.image.base import (_convert_patches_list_to_single_array,
                                  _create_patches_image)

    # If patches is a list, convert it to an array
    if isinstance(patches, list):
        patches = _convert_patches_list_to_single_array(patches,
                                                        patch_centers.n_points)

    # Create patches image
    if render_patches:
        patches_image = _create_patches_image(
            patches, patch_centers, patches_indices=patches_indices,
            offset_index=offset_index, background=background)
    else:
        if background == 'black':
            tmp_patches = np.zeros((patches.shape[0], patches.shape[1], 3,
                                    patches.shape[3], patches.shape[4]))
        elif background == 'white':
            tmp_patches = np.ones((patches.shape[0], patches.shape[1], 3,
                                   patches.shape[3], patches.shape[4]))
        patches_image = _create_patches_image(
            tmp_patches, patch_centers, patches_indices=patches_indices,
            offset_index=offset_index, background=background)
        channels = None

    # Render patches image
    if render_centers:
        patch_view = patches_image.view_landmarks(
            channels=channels, group='patch_centers', figure_id=figure_id,
            new_figure=new_figure, interpolation=interpolation,
            cmap_name=cmap_name, alpha=alpha, render_lines=render_lines,
            line_colour=line_colour, line_style=line_style,
            line_width=line_width, render_markers=render_markers,
            marker_style=marker_style, marker_size=marker_size,
            marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width,
            render_numbering=render_numbering,
            numbers_horizontal_align=numbers_horizontal_align,
            numbers_vertical_align=numbers_vertical_align,
            numbers_font_name=numbers_font_name,
            numbers_font_size=numbers_font_size,
            numbers_font_style=numbers_font_style,
            numbers_font_weight=numbers_font_weight,
            numbers_font_colour=numbers_font_colour,
            render_legend=False, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            axes_x_ticks=axes_x_ticks, axes_y_ticks=axes_y_ticks,
            figure_size=figure_size)
    else:
        patch_view = patches_image.view(
            figure_id=figure_id, new_figure=new_figure, channels=channels,
            interpolation=interpolation, cmap_name=cmap_name, alpha=alpha,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits, axes_x_ticks=axes_x_ticks,
            axes_y_ticks=axes_y_ticks, figure_size=figure_size)

    # Render rectangles around patches
    if render_patches_bboxes:
        patch_shape = [patches.shape[3], patches.shape[4]]
        render_rectangles_around_patches(
            patches_image.landmarks['patch_centers'].points,
            patch_shape, image_view=True, line_colour=bboxes_line_colour,
            line_style=bboxes_line_style, line_width=bboxes_line_width,
            interpolation=interpolation)

    return patch_view


def plot_gaussian_ellipses(covariances, means, n_std=2, render_colour_bar=True,
                           colour_bar_label='Normalized Standard Deviation',
                           colour_map='jet', figure_id=None, new_figure=False,
                           image_view=True, line_colour='r', line_style='-',
                           line_width=1., render_markers=True,
                           marker_edge_colour='k', marker_face_colour='k',
                           marker_edge_width=1., marker_size=5,
                           marker_style='o', render_axes=False,
                           axes_font_name='sans-serif', axes_font_size=10,
                           axes_font_style='normal', axes_font_weight='normal',
                           crop_proportion=0.1, figure_size=(7, 7)):
    r"""
    Method that renders the Gaussian ellipses that correspond to a set of
    covariance matrices and mean vectors. Naturally, this only works for
    2-dimensional random variables.

    Parameters
    ----------
    covariances : `list` of ``(2, 2)`` `ndarray`
        The covariance matrices that correspond to each ellipse.
    means : `list` of ``(2, )`` `ndarray`
        The mean vectors that correspond to each ellipse.
    n_std : `float`, optional
        This defines the size of the ellipses in terms of number of standard
        deviations.
    render_colour_bar : `bool`, optional
        If ``True``, then the ellipses will be coloured based on their
        normalized standard deviations and a colour bar will also appear on
        the side. If ``False``, then all the ellipses will have the same colour.
    colour_bar_label : `str`, optional
        The title of the colour bar. It only applies if `render_colour_bar`
        is ``True``.
    colour_map : `str`, optional
        A valid Matplotlib colour map. For more info, please refer to
        `matplotlib.cm`.
    figure_id : `object`, optional
        The id of the figure to be used.
    new_figure : `bool`, optional
        If ``True``, a new figure is created.
    image_view : `bool`, optional
        If ``True`` the ellipses will be rendered in the image coordinates
        system.
    line_colour : See Below, optional
        The colour of the lines of the ellipses.
        Example options::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    line_style : ``{-, --, -., :}``, optional
        The style of the lines of the ellipses.
    line_width : `float`, optional
        The width of the lines of the ellipses.
    render_markers : `bool`, optional
        If ``True``, the centers of the ellipses will be rendered.
    marker_style : See Below, optional
        The style of the centers of the ellipses. Example options ::

            {., ,, o, v, ^, <, >, +, x, D, d, s, p, *, h, H, 1, 2, 3, 4, 8}

    marker_size : `int`, optional
        The size of the centers of the ellipses in points.
    marker_face_colour : See Below, optional
        The face (filling) colour of the centers of the ellipses.
        Example options ::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    marker_edge_colour : See Below, optional
        The edge colour of the centers of the ellipses.
        Example options ::

            {r, g, b, c, m, k, w}
            or
            (3, ) ndarray

    marker_edge_width : `float`, optional
        The edge width of the centers of the ellipses.
    render_axes : `bool`, optional
        If ``True``, the axes will be rendered.
    axes_font_name : See Below, optional
        The font of the axes. Example options ::

            {serif, sans-serif, cursive, fantasy, monospace}

    axes_font_size : `int`, optional
        The font size of the axes.
    axes_font_style : ``{normal, italic, oblique}``, optional
        The font style of the axes.
    axes_font_weight : See Below, optional
        The font weight of the axes.
        Example options ::

            {ultralight, light, normal, regular, book, medium, roman,
            semibold,demibold, demi, bold, heavy, extra bold, black}

    crop_proportion : `float`, optional
        The proportion to be left around the centers' pointcloud.
    figure_size : (`float`, `float`) `tuple` or ``None`` optional
        The size of the figure in inches.
    """
    import matplotlib.pyplot as plt
    from matplotlib.patches import Ellipse
    import matplotlib.colors as colors
    import matplotlib.cm as cmx
    from matplotlib.font_manager import FontProperties
    from menpo.shape import PointCloud

    def eigh_sorted(cov):
        vals, vecs = np.linalg.eigh(cov)
        order = vals.argsort()[::-1]
        return vals[order], vecs[:, order]

    # get correct line style
    if line_style == '-':
        line_style = 'solid'
    elif line_style == '--':
        line_style = 'dashed'
    elif line_style == '-.':
        line_style = 'dashdot'
    elif line_style == ':':
        line_style = 'dotted'
    else:
        raise ValueError("line_style must be selected from "
                         "['-', '--', '-.', ':'].")

    # create pointcloud
    pc = PointCloud(np.array(means))

    # compute axes limits
    bounds = pc.bounds()
    r = pc.range()
    x_rr = r[0] * crop_proportion
    y_rr = r[1] * crop_proportion
    axes_x_limits = [bounds[0][1] - x_rr, bounds[1][1] + x_rr]
    axes_y_limits = [bounds[0][0] - y_rr, bounds[1][0] + y_rr]
    normalizer = np.sum(r) / 2.

    # compute height, width, theta and std
    stds = []
    heights = []
    widths = []
    thetas = []
    for cov in covariances:
        vals, vecs = eigh_sorted(cov)
        width, height = np.sqrt(vals)
        theta = np.degrees(np.arctan2(*vecs[:, 0][::-1]))
        stds.append(np.mean([height, width]) / normalizer)
        heights.append(height)
        widths.append(width)
        thetas.append(theta)

    if render_colour_bar:
        # set colormap values
        cmap = plt.get_cmap(colour_map)
        cNorm = colors.Normalize(vmin=np.min(stds), vmax=np.max(stds))
        scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=cmap)

    # visualize pointcloud
    if render_colour_bar:
        renderer = pc.view(
            figure_id=figure_id, new_figure=new_figure, image_view=image_view,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, axes_x_limits=axes_x_limits,
            axes_y_limits=axes_y_limits, figure_size=figure_size,
            render_markers=False)
    else:
        renderer = pc.view(
            figure_id=figure_id, new_figure=new_figure, image_view=image_view,
            marker_edge_colour=marker_edge_colour,
            marker_face_colour=marker_face_colour,
            marker_edge_width=marker_edge_width, marker_size=marker_size,
            marker_style=marker_style, render_axes=render_axes,
            axes_font_name=axes_font_name, axes_font_size=axes_font_size,
            axes_font_style=axes_font_style, axes_font_weight=axes_font_weight,
            axes_x_limits=axes_x_limits, axes_y_limits=axes_y_limits,
            figure_size=figure_size, render_markers=render_markers)

    # plot ellipses
    ax = plt.gca()
    for i in range(len(covariances)):
        # Width and height are "full" widths, not radius
        width = 2 * n_std * widths[i]
        height = 2 * n_std * heights[i]

        if image_view:
            colour = line_colour
            if render_colour_bar:
                colour = scalarMap.to_rgba(stds[i])
                if render_markers:
                    plt.plot(means[i][1], means[i][0], facecolor=colour,
                             edgecolor=colour, linewidth=0)
            ellip = Ellipse(xy=means[i][-1::-1], width=height, height=width,
                            angle=thetas[i], linestyle=line_style,
                            linewidth=line_width, edgecolor=colour,
                            facecolor='none')
        else:
            colour = line_colour
            if render_colour_bar:
                colour = scalarMap.to_rgba(stds[i])
                if render_markers:
                    plt.plot(means[i][0], means[i][1], facecolor=colour,
                             edgecolor=colour, linewidth=0)
            ellip = Ellipse(xy=means[i], width=width, height=height,
                            angle=thetas[i], linestyle=line_style,
                            linewidth=line_width, edgecolor=colour,
                            facecolor='none')
        ax.add_artist(ellip)

    # show colour bar
    if render_colour_bar:
        scalarMap.set_array(stds)
        cb = plt.colorbar(scalarMap, label=colour_bar_label)

        # change colour bar's font properties
        ax = cb.ax
        text = ax.yaxis.label
        font = FontProperties(size=axes_font_size, weight=axes_font_weight,
                              style=axes_font_style, family=axes_font_name)
        text.set_font_properties(font)

    return renderer



import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
from pytest import raises

from menpo.transform import Translation


def test_1d_translation():
    t_vec = np.array([1])
    with raises(ValueError):
        Translation(t_vec)


def test_5d_translation():
    t_vec = np.ones(5)
    with raises(ValueError):
        Translation(t_vec)


def test_translation():
    t_vec = np.array([1, 2, 3])
    starting_vector = np.random.rand(10, 3)
    transform = Translation(t_vec)
    transformed = transform.apply(starting_vector)
    assert_allclose(starting_vector + t_vec, transformed)


def test_translation_2d_from_vector():
    params = np.array([1, 2])
    homo = np.array([[1, 0, params[0]],
                     [0, 1, params[1]],
                     [0, 0, 1]])

    tr = Translation.init_identity(2).from_vector(params)

    assert_almost_equal(tr.h_matrix, homo)


def test_translation_2d_as_vector():
    params = np.array([1, 2])
    vec = Translation(params).as_vector()
    assert_allclose(vec, params)


def test_translation_3d_from_vector():
    params = np.array([1, 2, 3])
    homo = np.array([[1, 0, 0, params[0]],
                     [0, 1, 0, params[1]],
                     [0, 0, 1, params[2]],
                     [0, 0, 0, 1]])

    tr = Translation.init_identity(3).from_vector(params)

    assert_almost_equal(tr.h_matrix, homo)


def test_translation_3d_as_vector():
    params = np.array([1, 2, 3])
    vec = Translation(params).as_vector()
    assert_allclose(vec, params)


def test_translation_2d_n_parameters():
    trans = np.array([1, 2])
    t = Translation(trans)
    assert (t.n_parameters == 2)


def test_translation_3d_n_parameters():
    trans = np.array([1, 2, 3])
    t = Translation(trans)
    assert (t.n_parameters == 3)


def test_translation_from_list():
    t_a = Translation([3, 4])
    t_b = Translation(np.array([3, 4]))
    assert (np.all(t_a.h_matrix == t_b.h_matrix))


def test_translation_identity_2d():
    assert_allclose(Translation.init_identity(2).h_matrix, np.eye(3))


def test_translation_identity_3d():
    assert_allclose(Translation.init_identity(3).h_matrix, np.eye(4))


def test_translation_decompose_optional():
    t = Translation.init_identity(2)
    d = t.decompose()
    assert np.all(d[0].h_matrix == t.h_matrix)



import numpy as np
import cython
cimport numpy as np
cimport cython
from libc.math cimport sqrt
from menpo.cy_utils cimport dtype_from_memoryview


ctypedef fused floats:
    np.float32_t
    np.float64_t

ctypedef fused integrals:
    np.int16_t
    np.uint16_t
    np.uint32_t
    np.uint64_t
    np.int32_t
    np.int64_t


cdef normalize(floats[:, :] vec):
    """
    Normalize the given array of vectors inplace.

    Parameters
    ----------
    vec : (N, 3) c-contiguous float32/float64 ndarray
        The array of vectors to normalize
    """
    # Avoid divisions by almost 0 numbers
    # np.spacing(1) is equivalent to Matlab's eps
    cdef:
        floats mag
        Py_ssize_t n_elems = vec.shape[0], i = 0
        floats eps = np.spacing(1)

    for i in range(n_elems):
        mag = sqrt(vec[i, 0] * vec[i, 0] +
                   vec[i, 1] * vec[i, 1] +
                   vec[i, 2] * vec[i, 2])
        # Zero magnitude (or close to) causes massive values, so just set vector
        # to zero.
        if mag < eps:
            vec[i, 0] = 0.0
            vec[i, 1] = 0.0
            vec[i, 2] = 0.0
        else:
            vec[i, 0] /= mag
            vec[i, 1] /= mag
            vec[i, 2] /= mag


cdef inline floats[:, :] triangle_cross(floats[:, :] vertex,
                                        integrals[:, :] face):
    """
    The N x 3 cross product of the two sides of the triangles defined
    by the face array.

    Parameters
    ----------
    x : (N, 3) float32/float64 memory view
        First array to perform cross product with.
    y : (N, 3) float32/float64 memory view
        Second array to perform cross product with.

    Returns
    -------
    cross : (N, 3) c-contiguous float32/float64 ndarray
        The array of vectors representing the cross product between each
        corresponding vector.
    """
    vertex_dtype = dtype_from_memoryview(vertex)
    cdef:
        Py_ssize_t n_vert = vertex.shape[0], n_face = face.shape[0]
        Py_ssize_t i = 0, j = 0
        floats[:, :] z = np.zeros([n_face, 3], dtype=vertex_dtype)
        floats *v0
        floats *v1
        floats *v2
        floats x[3]
        floats y[3]

    for i in range(n_face):
        # This is dangerous since we just have a raw pointer, but since we know
        # our vectors are always [n, 3] we hard coded the 3 loop below.
        v0 = &vertex[face[i, 0], :][0]
        v1 = &vertex[face[i, 1], :][0]
        v2 = &vertex[face[i, 2], :][0]
        # Calculate triangle edge vectors
        for j in range(3):
            x[j] = v1[j] - v0[j]
            y[j] = v2[j] - v0[j]
        # Cross Product
        z[i, 0] = x[1] * y[2] - x[2] * y[1]
        z[i, 1] = x[2] * y[0] - x[0] * y[2]
        z[i, 2] = x[0] * y[1] - x[1] * y[0]

    return z


cpdef compute_vertex_normals(floats[:, :] vertex, integrals[:, :] face):
    """
    Compute the per-vertex normals of the vertices given a list of
    faces.

    Parameters
    ----------
    vertex : (N, 3) c-contiguous float32/float64 ndarray
        The list of points to compute normals for.
    face : (M, 3) c-contiguous int16/int32/int64 ndarray
        The list of faces (triangle list).

    Returns
    -------
    vertex_normal : (N, 3) c-contiguous float32/float64 ndarray
        The normal per vertex.
    :return:
    """
    cdef:
        Py_ssize_t i = 0, j = 0
        Py_ssize_t n_vert = vertex.shape[0]
        Py_ssize_t n_face = face.shape[0]
        floats[:, :] face_normal
        floats[:, :] vertex_normal = np.zeros_like(vertex)
        integrals f0, f1, f2

    # Calculate the cross product (per-face normal)
    face_normal = triangle_cross(vertex, face)
    normalize(face_normal)

    # Calculate per-vertex normal
    for i in range(n_face):
        f0 = face[i, 0]
        f1 = face[i, 1]
        f2 = face[i, 2]
        for j in range(3):
            vertex_normal[f0, j] += face_normal[i, j]
            vertex_normal[f1, j] += face_normal[i, j]
            vertex_normal[f2, j] += face_normal[i, j]
    # Normalize
    normalize(vertex_normal)

    return np.asarray(vertex_normal)


cpdef compute_face_normals(floats[:, :] vertex, integrals[:, :] face):
    """
    Compute per-face normals of the vertices given a list of
    faces.

    Parameters
    ----------
    vertex : (N, 3) c-contiguous float32/float64 ndarray
        The list of points to compute normals for.
    face : (M, 3) c-contiguous int16/int32/int64 ndarray
        The list of faces (triangle list).

    Returns
    -------
    face_normal : (M, 3) c-contiguous float32/float64 ndarray
        The normal per face.
    :return:
    """
    # Calculate the cross product (per-face normal)
    cdef floats[:, :] face_normal = triangle_cross(vertex, face)
    normalize(face_normal)

    return np.asarray(face_normal)



class MeshImportError(Exception):
    r"""
    Raised when a mesh importer finds an unexpected data format.
    """
    pass


class OverwriteError(ValueError):
    r"""
    Raised when an IO action would lead to the overwriting of an existing file
    without explit intention.
    """
    def __init__(self, message, path, *args):
        self.message = message  # without this you may get DeprecationWarning
        self.path = path
        # allow users initialize misc. arguments as any other builtin Error
        super(OverwriteError, self).__init__(message, *args)



import numpy as np
from numpy.testing import assert_allclose
from menpo.image import BooleanImage
from menpo.shape import PointCloud


def test_boolean_image_constrain_landmarks():
    mask = BooleanImage.init_blank((10, 10), fill=False)
    mask.landmarks['test'] = PointCloud(
        np.array([[1, 1], [8, 1], [8, 8], [1, 8]]))
    new_mask = mask.constrain_to_landmarks('test')
    assert_allclose(new_mask.pixels[1:-1, 1:-1], True)
    assert new_mask.n_true() == 64


def test_boolean_image_constrain_pointcloud_pwa():
    mask = BooleanImage.init_blank((10, 10), fill=False)
    pc = PointCloud(np.array([[1, 1], [8, 1], [8, 8], [1, 8]]))
    new_mask = mask.constrain_to_pointcloud(pc, point_in_pointcloud='pwa')
    assert_allclose(new_mask.pixels[:, 1:-1, 1:-1], True)
    assert new_mask.n_true() == 64


def test_boolean_image_constrain_pointcloud_convex_hull():
    mask = BooleanImage.init_blank((10, 10), fill=False)
    pc = PointCloud(np.array([[1, 1], [8, 1], [8, 8], [1, 8]]))
    new_mask = mask.constrain_to_pointcloud(pc,
                                            point_in_pointcloud='convex_hull')
    assert_allclose(new_mask.pixels[:, 2:-1, 2:-1], True)
    # Points on the boundary are OUTSIDE
    assert new_mask.n_true() == 56


def test_boolean_image_init_from_pointcloud():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = BooleanImage.init_from_pointcloud(pc, fill=False, constrain=False)
    assert im.n_true() == 0
    assert im.shape == (15, 15)


def test_boolean_image_init_from_pointcloud_constraint():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = BooleanImage.init_from_pointcloud(pc, fill=False, constrain=True)
    assert im.n_true() == 120
    assert im.shape == (15, 15)


def test_boolean_image_init_from_pointcloud_constrain_all_true():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = BooleanImage.init_from_pointcloud(pc, fill=True, constrain=True)
    assert im.n_true() == 120
    assert im.shape == (15, 15)



import warnings
import numpy as np
from numpy.testing import assert_allclose
from menpo.image import Image, MaskedImage
from menpo.shape import TriMesh, TexturedTriMesh, ColouredTriMesh
from menpo.testing import is_same_array


def test_trimesh_creation():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    TriMesh(points, trilist=trilist)


def test_trimesh_init_2d_grid():
    tm = TriMesh.init_2d_grid([10, 10])
    assert tm.n_points == 100
    assert tm.n_dims == 2
    # 162 = 9 * 9 * 2
    assert_allclose(tm.trilist.shape, (162, 3))
    assert_allclose(tm.range(), [9, 9])


def test_trimesh_init_from_depth_image():
    fake_z = np.random.uniform(size=(10, 10))
    tm = TriMesh.init_from_depth_image(Image(fake_z))
    assert tm.n_points == 100
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [9, 9])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_trimesh_init_from_depth_image_masked():
    fake_z = np.random.uniform(size=(10, 10))
    mask = np.zeros(fake_z.shape, dtype=np.bool)
    mask[2:6, 2:6] = True
    im = MaskedImage(fake_z, mask=mask)
    tm = TriMesh.init_from_depth_image(im)
    assert tm.n_points == 16
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [3, 3])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_colouredtrimesh_init_2d_grid():
    tm = ColouredTriMesh.init_2d_grid([10, 10])
    assert tm.n_points == 100
    assert tm.n_dims == 2
    # 162 = 9 * 9 * 2
    assert_allclose(tm.trilist.shape, (162, 3))
    assert_allclose(tm.range(), [9, 9])


def test_colouredtrimesh_init_from_depth_image():
    fake_z = np.random.uniform(size=(10, 10))
    tm = ColouredTriMesh.init_from_depth_image(Image(fake_z))
    assert tm.n_points == 100
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [9, 9])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_colouredtrimesh_init_from_depth_image_masked():
    fake_z = np.random.uniform(size=(10, 10))
    mask = np.zeros(fake_z.shape, dtype=np.bool)
    mask[2:6, 2:6] = True
    im = MaskedImage(fake_z, mask=mask)
    tm = ColouredTriMesh.init_from_depth_image(im)
    assert tm.n_points == 16
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [3, 3])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_colouredtrimesh_init_from_depth_image_coloured():
    fake_z = np.random.uniform(size=(10, 10))
    fake_colours = np.random.uniform(size=(100, 3))
    tm = ColouredTriMesh.init_from_depth_image(Image(fake_z),
                                               colours=fake_colours)
    assert tm.n_points == 100
    assert tm.n_dims == 3
    assert tm.colours.shape == (100, 3)
    assert_allclose(tm.range()[:2], [9, 9])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_texturedtrimesh_init_2d_grid():
    tm = TexturedTriMesh.init_2d_grid([10, 10])
    assert tm.n_points == 100
    assert tm.n_dims == 2
    # 162 = 9 * 9 * 2
    assert_allclose(tm.trilist.shape, (162, 3))
    assert_allclose(tm.range(), [9, 9])


def test_texturedtrimesh_init_from_depth_image():
    fake_z = np.random.uniform(size=(10, 10))
    tm = TexturedTriMesh.init_from_depth_image(Image(fake_z))
    assert tm.n_points == 100
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [9, 9])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_texturedtrimesh_init_from_depth_image_masked():
    fake_z = np.random.uniform(size=(10, 10))
    mask = np.zeros(fake_z.shape, dtype=np.bool)
    mask[2:6, 2:6] = True
    im = MaskedImage(fake_z, mask=mask)
    tm = TexturedTriMesh.init_from_depth_image(im)
    assert tm.n_points == 16
    assert tm.n_dims == 3
    assert_allclose(tm.range()[:2], [3, 3])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_texturedtrimesh_init_from_depth_image_textured():
    fake_z = np.random.uniform(size=(10, 10))
    fake_tex = Image(fake_z)
    tm = TexturedTriMesh.init_from_depth_image(fake_tex, texture=fake_tex)
    assert tm.n_points == 100
    assert tm.n_dims == 3
    assert tm.texture.shape == (10, 10)
    assert_allclose(tm.range()[:2], [9, 9])
    assert tm.points[:, -1].max() <= 1.0
    assert tm.points[:, -1].min() >= 0.0


def test_trimesh_creation_copy_true():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    tm = TriMesh(points, trilist=trilist)
    assert (not is_same_array(tm.points, points))
    assert (not is_same_array(tm.trilist, trilist))


def test_trimesh_creation_copy_false():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    tm = TriMesh(points, trilist=trilist, copy=False)
    assert (is_same_array(tm.points, points))
    assert (is_same_array(tm.trilist, trilist))


def test_texturedtrimesh_creation_copy_false():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    pixels = np.ones([10, 10])
    tcoords = np.ones([4, 2])
    texture = Image(pixels, copy=False)
    ttm = TexturedTriMesh(points, tcoords, texture, trilist=trilist,
                          copy=False)
    assert (is_same_array(ttm.points, points))
    assert (is_same_array(ttm.trilist, trilist))
    assert (is_same_array(ttm.tcoords.points, tcoords))
    assert (is_same_array(ttm.texture.pixels, pixels))


def test_texturedtrimesh_creation_copy_true():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    pixels = np.ones([10, 10, 1])
    tcoords = np.ones([4, 2])
    texture = Image(pixels, copy=False)
    ttm = TexturedTriMesh(points, tcoords, texture, trilist=trilist,
                          copy=True)
    assert (not is_same_array(ttm.points, points))
    assert (not is_same_array(ttm.trilist, trilist))
    assert (not is_same_array(ttm.tcoords.points, tcoords))
    assert (not is_same_array(ttm.texture.pixels, pixels))


def test_colouredtrimesh_creation_copy_false():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    colours = np.ones([4, 13])
    ttm = ColouredTriMesh(points, trilist=trilist, colours=colours, copy=False)
    assert (is_same_array(ttm.points, points))
    assert (is_same_array(ttm.trilist, trilist))
    assert (is_same_array(ttm.colours, colours))


def test_colouredtrimesh_creation_copy_true():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    colours = np.ones([4, 13])
    ttm = ColouredTriMesh(points, trilist=trilist, colours=colours, copy=True)
    assert (not is_same_array(ttm.points, points))
    assert (not is_same_array(ttm.trilist, trilist))
    assert (not is_same_array(ttm.colours, colours))


def test_trimesh_creation_copy_warning():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]], order='F')
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        TriMesh(points, trilist=trilist, copy=False)
        assert len(w) == 1


def test_trimesh_n_dims():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    trimesh = TriMesh(points, trilist=trilist)
    assert(trimesh.n_dims == 3)


def test_trimesh_n_points():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    trimesh = TriMesh(points, trilist=trilist)
    assert(trimesh.n_points == 4)


def test_trimesh_n_tris():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    trimesh = TriMesh(points, trilist=trilist)
    assert(trimesh.n_tris == 2)


def test_trimesh_from_tri_mask():
    points = np.array([[0, 0, 0],
                       [1, 0, 0],
                       [1, 1, 0],
                       [0, 1, 0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    mask = np.zeros(2, dtype=np.bool)
    mask[0] = True
    trimesh = TriMesh(points, trilist=trilist).from_tri_mask(mask)
    assert(trimesh.n_tris == 1)
    assert(trimesh.n_points == 3)
    assert_allclose(trimesh.points, points[trilist[0]])


def test_trimesh_face_normals():
    points = np.array([[0.0, 0.0, -1.0],
                       [1.0, 0.0, 0.0],
                       [1.0, 1.0, 0.0],
                       [0.0, 1.0, 0.0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    expected_normals = np.array([[-np.sqrt(3)/3, -np.sqrt(3)/3, np.sqrt(3)/3],
                                 [-0, -0, 1]])
    trimesh = TriMesh(points, trilist=trilist)
    face_normals = trimesh.tri_normals()
    assert_allclose(face_normals, expected_normals)

    trimesh.points = trimesh.points.astype(np.float32)
    face_normals = trimesh.tri_normals()
    assert_allclose(face_normals, expected_normals)


def test_trimesh_vertex_normals():
    points = np.array([[0.0, 0.0, -1.0],
                       [1.0, 0.0, 0.0],
                       [1.0, 1.0, 0.0],
                       [0.0, 1.0, 0.0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    # 0 and 2 are the corner of the triangles and so the maintain the
    # face normals. The other two are the re-normalized vertices:
    # normalize(n0 + n2)
    expected_normals = np.array([[-np.sqrt(3)/3, -np.sqrt(3)/3, np.sqrt(3)/3],
                                 [-0.32505758,  -0.32505758, 0.88807383],
                                 [0, 0, 1],
                                 [-0.32505758,  -0.32505758, 0.88807383]])
    trimesh = TriMesh(points, trilist)
    vertex_normals = trimesh.vertex_normals()
    assert_allclose(vertex_normals, expected_normals)

    trimesh.points = trimesh.points.astype(np.float32)
    vertex_normals = trimesh.vertex_normals()
    assert_allclose(vertex_normals, expected_normals)


def test_cython_face_normals_same_as_pure_python():
    from .. import normals, pynormals
    points = np.array([[0.0, 0.0, -1.0],
                       [1.0, 0.0, 0.0],
                       [1.0, 1.0, 0.0],
                       [0.0, 1.0, 0.0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    assert_allclose(normals.compute_face_normals(points, trilist),
                    pynormals.compute_face_normals(points, trilist))


def test_cython_vertex_normals_same_as_pure_python():
    from .. import normals, pynormals
    points = np.array([[0.0, 0.0, -1.0],
                       [1.0, 0.0, 0.0],
                       [1.0, 1.0, 0.0],
                       [0.0, 1.0, 0.0]])
    trilist = np.array([[0, 1, 3],
                        [1, 2, 3]])
    assert_allclose(normals.compute_vertex_normals(points, trilist),
                    pynormals.compute_vertex_normals(points, trilist))



import menpo.io as mio
from menpo.base import copy_landmarks_and_path
from menpo.image import Image


def test_copy_landmarks_and_path():
    img = mio.import_builtin_asset.lenna_png()
    new_img = Image.init_blank(img.shape)
    copy_landmarks_and_path(img, new_img)

    assert new_img.path == img.path
    assert new_img.landmarks.keys() == img.landmarks.keys()
    assert new_img.landmarks is not img.landmarks


def test_copy_landmarks_and_path_returns_target():
    img = mio.import_builtin_asset.lenna_png()
    new_img = Image.init_blank(img.shape)
    new_img_ret = copy_landmarks_and_path(img, new_img)
    assert new_img_ret is new_img


def test_copy_landmarks_and_path_with_no_lms_path():
    img = Image.init_blank((5, 5))
    new_img = Image.init_blank((5, 5))
    copy_landmarks_and_path(img, new_img)
    assert not hasattr(img, 'path')
    assert not hasattr(new_img, 'path')
    assert not img.has_landmarks
    assert not new_img.has_landmarks



from menpo.base import partial_doc

from .landmark import asf_importer, pts_importer


asf_image_importer = partial_doc(asf_importer, image_origin=True)

pts_image_importer = partial_doc(pts_importer, image_origin=True)



import menpo


def test_image_gaussian_pyramid_n_levels():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    assert len(list(lenna.gaussian_pyramid(n_levels=4))) == 4


def test_image_gaussian_pyramid_one_level():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    assert len(list(lenna.gaussian_pyramid(n_levels=1))) == 1


def test_image_gaussian_pyramid_shapes():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    shapes = [(512, 512), (256, 256), (128, 128), (64, 64)]
    for l, expected_shape in zip(lenna.gaussian_pyramid(n_levels=4), shapes):
        assert l.shape == expected_shape


def test_image_pyramid_n_levels():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    assert len(list(lenna.pyramid(n_levels=4))) == 4


def test_image_pyramid_one_level():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    assert len(list(lenna.pyramid(n_levels=1))) == 1


def test_image_pyramid_shapes():
    lenna = menpo.io.import_builtin_asset.lenna_png()
    shapes = [(512, 512), (128, 128), (32, 32)]
    for l, expected_shape in zip(lenna.pyramid(n_levels=3, downscale=4), shapes):
        assert l.shape == expected_shape



from collections import OrderedDict
import numpy as np

from ..base import (labeller_func, validate_input, connectivity_from_array,
                    pcloud_and_lgroup_from_ranges)


@labeller_func(group_label='pose_stickmen_12')
def pose_stickmen_12_to_pose_stickmen_12(pcloud):
    r"""
    Apply the 'stickmen' 12-point semantic labels.

    The semantic labels applied are as follows:

      - torso
      - right_upper_arm
      - left_upper_arm
      - right_lower_arm
      - left_lower_arm
      - head

    References
    ----------
    .. [1] http://www.robots.ox.ac.uk/~vgg/data/stickmen/
    """
    n_expected_points = 12
    validate_input(pcloud, n_expected_points)

    labels = OrderedDict([
        ('torso', (0, 2, False)),
        ('right_upper arm', (2, 4, False)),
        ('left_upper arm', (4, 6, False)),
        ('right_lower_arm', (6, 8, False)),
        ('left_lower_arm', (8, 10, False)),
        ('head', (10, 12, False))
    ])
    return pcloud_and_lgroup_from_ranges(pcloud, labels)


@labeller_func(group_label='pose_lsp_14')
def pose_lsp_14_to_pose_lsp_14(pcloud):
    r"""
    Apply the lsp 14-point semantic labels.

    The semantic labels applied are as follows:

      - left_leg
      - right_leg
      - left_arm
      - right_arm
      - head

    References
    ----------
    .. [1] http://www.comp.leeds.ac.uk/mat4saj/lsp.html
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 14
    validate_input(pcloud, n_expected_points)

    left_leg_indices = np.arange(0, 3)
    right_leg_indices = np.arange(3, 6)
    left_arm_indices = np.arange(6, 9)
    right_arm_indices = np.arange(9, 12)
    head_indices = np.arange(12, 14)

    left_leg_connectivity = connectivity_from_array(left_leg_indices)
    right_leg_connectivity = connectivity_from_array(right_leg_indices)
    left_arm_connectivity = connectivity_from_array(left_arm_indices)
    right_arm_connectivity = connectivity_from_array(right_arm_indices)
    head_connectivity = connectivity_from_array(head_indices)

    all_connectivity = np.vstack([
        left_leg_connectivity, right_leg_connectivity,
        left_arm_connectivity, right_arm_connectivity,
        head_connectivity
    ])

    mapping = OrderedDict()
    mapping['left_leg'] = left_leg_indices
    mapping['right_leg'] = right_leg_indices
    mapping['left_arm'] = left_arm_indices
    mapping['right_arm'] = right_arm_indices
    mapping['head'] = head_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping)

    return new_pcloud, mapping


@labeller_func(group_label='pose_flic_11')
def pose_flic_11_to_pose_flic_11(pcloud):
    r"""
    Apply the flic 11-point semantic labels.

    The semantic labels applied are as follows:

      - left_arm
      - right_arm
      - hips
      - face

    References
    ----------
    .. [1] http://vision.grasp.upenn.edu/cgi-bin/index.php?n=VideoLearning.FLIC
    """
    n_expected_points = 11
    validate_input(pcloud, n_expected_points)

    labels = OrderedDict([
        ('left_arm', (0, 3, False)),
        ('right_arm', (3, 6, False)),
        ('hips', (6, 8, False)),
        ('face', (8, 11, True))])

    return pcloud_and_lgroup_from_ranges(pcloud, labels)


@labeller_func(group_label='pose_human36M_32')
def pose_human36M_32_to_pose_human36M_32(pcloud):
    r"""
    Apply the human3.6M 32-point semantic labels.

    The semantic labels applied are as follows:

      - pelvis
      - right_leg
      - left_leg
      - spine
      - head
      - left_arm
      - left_hand
      - right_arm
      - right_hand
      - torso

    References
    ----------
    .. [1] http://vision.imar.ro/human3.6m/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 32
    validate_input(pcloud, n_expected_points)

    pelvis_indices = np.array([1, 0, 6])
    right_leg_indices = np.array(range(1, 6))
    left_leg_indices = np.array(range(6, 11))
    spine_indices = np.array([11, 12, 13])
    head_indices = np.array([13, 14, 15])
    left_arm_indices = np.array([16, 17, 18, 19, 23])
    left_hand_indices = np.array([20, 21, 22])
    right_arm_indices = np.array([24, 25, 26, 27, 29, 31])
    right_hand_indices = np.array([28, 29, 30])
    torso_indices = np.array([0, 1, 25, 13, 17, 6])

    pelvis_connectivity = connectivity_from_array(pelvis_indices)
    right_leg_connectivity = connectivity_from_array(right_leg_indices)
    left_leg_connectivity = connectivity_from_array(left_leg_indices)
    spine_connectivity = connectivity_from_array(spine_indices)
    head_connectivity = connectivity_from_array(head_indices)
    left_arm_connectivity = connectivity_from_array(left_arm_indices)
    left_hand_connectivity = connectivity_from_array(left_hand_indices)
    right_arm_connectivity = connectivity_from_array(right_arm_indices)
    right_hand_connectivity = connectivity_from_array(right_hand_indices)
    torso_connectivity = connectivity_from_array(torso_indices,
                                                 close_loop=True)

    all_connectivity = np.vstack([
        pelvis_connectivity, right_leg_connectivity, left_leg_connectivity,
        spine_connectivity, head_connectivity, left_arm_connectivity,
        left_hand_connectivity, right_arm_connectivity,
        right_hand_connectivity, torso_connectivity
    ])

    mapping = OrderedDict()
    mapping['pelvis'] = pelvis_indices
    mapping['right_leg'] = right_leg_indices
    mapping['left_leg'] = left_leg_indices
    mapping['spine'] = spine_indices
    mapping['head'] = head_indices
    mapping['left_arm'] = left_arm_indices
    mapping['left_hand'] = left_hand_indices
    mapping['right_arm'] = right_arm_indices
    mapping['right_hand'] = right_hand_indices
    mapping['torso'] = torso_indices

    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points, all_connectivity, mapping)

    return new_pcloud, mapping


@labeller_func(group_label='pose_human36M_17')
def pose_human36M_32_to_pose_human36M_17(pcloud):
    r"""
    Apply the human3.6M 17-point semantic labels (based on the
    original semantic labels of Human3.6 but removing the annotations
    corresponding to duplicate points, soles and palms), originally 32-points.

    The semantic labels applied are as follows:

      - pelvis
      - right_leg
      - left_leg
      - spine
      - head
      - left_arm
      - right_arm
      - torso

    References
    ----------
    .. [1] http://vision.imar.ro/human3.6m/
    """
    from menpo.shape import LabelledPointUndirectedGraph

    n_expected_points = 32
    validate_input(pcloud, n_expected_points)

    pelvis_indices = np.array([1, 0, 4])
    right_leg_indices = np.arange(1, 4)
    left_leg_indices = np.arange(4, 7)
    spine_indices = np.array([0, 7, 8])
    head_indices = np.array([8, 9, 10])
    left_arm_indices = np.array([8, 11, 12, 13])
    right_arm_indices = np.array([8, 14, 15, 16])
    torso_indices = np.array([0, 1, 14, 8, 11, 4])

    pelvis_connectivity = connectivity_from_array(pelvis_indices)
    right_leg_connectivity = connectivity_from_array(right_leg_indices)
    left_leg_connectivity = connectivity_from_array(left_leg_indices)
    spine_connectivity = connectivity_from_array(spine_indices)
    head_connectivity = connectivity_from_array(head_indices)
    left_arm_connectivity = connectivity_from_array(left_arm_indices)
    right_arm_connectivity = connectivity_from_array(right_arm_indices)
    torso_connectivity = connectivity_from_array(torso_indices,
                                                 close_loop=True)

    all_connectivity = np.vstack([
        pelvis_connectivity, right_leg_connectivity, left_leg_connectivity,
        spine_connectivity, head_connectivity, left_arm_connectivity,
        right_arm_connectivity, torso_connectivity
    ])

    mapping = OrderedDict()
    mapping['pelvis'] = pelvis_indices
    mapping['right_leg'] = right_leg_indices
    mapping['left_leg'] = left_leg_indices
    mapping['spine'] = spine_indices
    mapping['head'] = head_indices
    mapping['left_arm'] = left_arm_indices
    mapping['right_arm'] = right_arm_indices
    mapping['torso'] = torso_indices

    # Ignore duplicate points, sole and palms
    ind = np.hstack([np.arange(0, 4), np.arange(6, 9), np.arange(12, 16),
                     np.arange(17, 20), np.arange(25, 28)])
    new_pcloud = LabelledPointUndirectedGraph.init_from_indices_mapping(
        pcloud.points[ind], all_connectivity, mapping)

    return new_pcloud, mapping



from __future__ import division
from menpo.shape import PointCloud


def mean_pointcloud(pointclouds):
    r"""
    Compute the mean of a `list` of :map:`PointCloud` or subclass objects.
    The list is assumed to be homogeneous i.e all elements of the list are
    assumed to belong to the same point cloud subclass just as all elements
    are also assumed to have the same number of points and represent
    semantically equivalent point clouds.

    Parameters
    ----------
    pointclouds: `list` of :map:`PointCloud` or subclass
        List of point cloud or subclass objects from which we want to compute
        the mean.

    Returns
    -------
    mean_pointcloud : :map:`PointCloud` or subclass
        The mean point cloud or subclass.
    """
    # make a temporary PointCloud (with copy=False for low overhead)
    tmp_pc = PointCloud(sum(pc.points for pc in pointclouds) /
                        len(pointclouds), copy=False)
    # use the type of the first element in the list to rebuild from the vector
    return pointclouds[0].from_vector(tmp_pc.as_vector())



import warnings
import os
import numpy as np
import subprocess as sp
import re
from pathlib import Path

from menpo.image.base import normalize_pixels_range, channels_to_front
from menpo.image import Image
from menpo.base import LazyList

from ..utils import DEVNULL, _call_subprocess


_FFMPEG_CMD = lambda: str(Path(os.environ.get('MENPO_FFMPEG_CMD', 'ffmpeg')))
_FFPROBE_CMD = lambda: str(Path(os.environ.get('MENPO_FFPROBE_CMD', 'ffprobe')))


def ffmpeg_importer(filepath, normalize=True, exact_frame_count=True, **kwargs):
    r"""
    Imports videos by streaming frames from a pipe using FFMPEG. Returns a
    :map:`LazyList` that gives lazy access to the video on a per-frame basis.

    There are two important environment variables that can be set to alter
    the behaviour of this function:

        ================== ======================================
        ENV Variable       Definition
        ================== ======================================
        MENPO_FFMPEG_CMD   The path to the 'ffmpeg' executable.
        MENPO_FFPROBE_CMD  The path to the 'ffprobe' executable.
        ================== ======================================

    Parameters
    ----------
    filepath : `Path`
        Absolute filepath of the video.
    normalize : `bool`, optional
        If ``True``, normalize between 0.0 and 1.0 and convert to float. If
        ``False`` just return whatever ffmpeg imports.
    exact_frame_count: `bool`, optional
        If ``True``, the import fails if ffprobe is not available
        (reading from ffmpeg's output returns inexact frame count)
    \**kwargs : `dict`, optional
        Any other keyword arguments.

    Returns
    -------
    image : :map:`LazyList`
        A :map:`LazyList` containing :map:`Image` or subclasses per frame
        of the video.
    """
    reader = FFMpegVideoReader(filepath, normalize=normalize, exact_frame_count=exact_frame_count)
    ll = LazyList.init_from_index_callable(lambda x: Image.init_from_channels_at_back(reader[x]), len(reader))
    ll.fps = reader.fps

    return ll


def ffmpeg_types():
    r"""The supported FFMPEG types.

    Returns
    -------
    supported_types : `dict`
        A dictionary of extensions supported by the FFMPEG importer.
    """
    # Need to check this
    ffmpeg_exts = ['.avi', '.mp4', '.mpg', '.mpeg', '.wmv', '.mov', '.mkv']
    return {ext: ffmpeg_importer for ext in ffmpeg_exts}


class FFMpegVideoReader(object):
    """
    Read a video using ffmpeg and handle state to allow seeking.

    Parameters
    ----------
    filepath : `Path`
        Absolute path to the video
    normalize : `bool`, optional
        If ``True``, the resulting range of the pixels of the returned
        frames is normalized.
    exact_frame_count : `bool`, optional
        If True, the import fails if ffmprobe is not available
        (reading from ffmpeg's output returns inexact frame count)
    """
    def __init__(self, filepath, normalize=False, exact_frame_count=True):
        self.filepath = filepath
        self.normalize = normalize
        self.exact_frame_count = exact_frame_count
        self._pipe = None
        if self.exact_frame_count:
            try:
                infos = video_infos_ffprobe(self.filepath)
            except:
                raise ValueError('ffprobe not available, unable to get exact '
                                 'frame count. If you want to use an '
                                 'approximate frame number, set exact_frame_'
                                 'count to False and proceed at your own '
                                 'risk.')
        else:
            infos = video_infos_ffmpeg(self.filepath)
        self.duration = infos['duration']
        self.width = infos['width']
        self.height = infos['height']
        self.n_frames = infos['n_frames']
        self.fps = infos['fps']
        # contains the index of the last read frame
        # the index is updated in _open_pipe, _read_one_frame and _trash_frames
        self.index = -1

    def _shutdown_pipe(self):
        if self._pipe is not None:
            if self._pipe.stdout:
                self._pipe.stdout.close()
            if self._pipe.stderr:
                self._pipe.stderr.close()
            if self._pipe.stdin:
                self._pipe.stdin.close()
        self._pipe = None

    def __del__(self):
        r"""
        Close the pipe if open.
        """
        self._shutdown_pipe()

    def __len__(self):
        return self.n_frames

    def _open_pipe(self, frame=None):
        r"""
        Open a pipe at the time just before the specified frame

        Parameters
        ----------
        frame : `int`, optional
            If ``None``, pipe opened from the beginning of the video
            otherwise, pipe opened at the time corresponding to that frame

        Note
        ----
        Since v.2.1 of ffmpeg, this is frame-accurate
        """
        if frame is not None and frame > 0:
            time = str(frame / float(self.fps))
            command = [_FFMPEG_CMD(),
                       '-ss', time,
                       '-i', str(self.filepath),
                       '-f', 'image2pipe',
                       '-pix_fmt', 'rgb24',
                       '-vcodec', 'rawvideo', '-']
        else:
            command = [_FFMPEG_CMD(),
                       '-i', str(self.filepath),
                       '-f', 'image2pipe',
                       '-pix_fmt', 'rgb24',
                       '-vcodec', 'rawvideo', '-']
            frame = 0

        self._shutdown_pipe()
        self._pipe = sp.Popen(command, stdout=sp.PIPE, stdin=DEVNULL,
                              stderr=DEVNULL,
                              bufsize=10**8)  # Is this buffer the correct size?
        # We have not yet read the specified frame
        self.index = frame - 1

    def __iter__(self):
        r"""
        Iterate through all frames of the video in order

        Only opens the pipe once at the beginning
        """
        self.index = 0
        for index in range(self.n_frames):
            yield self[index]

    def __getitem__(self, index):
        r"""
        Get a specific frame from the video
        """
        # If the user is reading consecutive frames, or a frame later in the
        # video, do not reopen a pipe
        if self._pipe is None or self._pipe.poll() is not None or index <= self.index:
            self._open_pipe(frame=index)
        else:
            to_trash = index - self.index - 1
            if to_trash > 0:
                self._trash_frames(to_trash)

        return self._read_one_frame()

    def _trash_frames(self, n_frames):
        r"""
        Reads and trashes the data corresponding to ``n_frames``
        """
        _ = self._pipe.stdout.read(self.height*self.width*3*n_frames)
        self._pipe.stdout.flush()
        self.index += n_frames

    def _read_one_frame(self):
        r"""
        Reads one frame from the opened ``self._pipe`` and converts it to
        a numpy array

        Returns
        -------
        image : :map:`Image`
            Image of shape ``(self.height, self.width, 3)``
        """
        raw_data = self._pipe.stdout.read(self.height*self.width*3)
        frame = np.frombuffer(raw_data, dtype=np.uint8)
        frame = frame.reshape((self.height, self.width, 3))
        self._pipe.stdout.flush()
        self.index += 1

        if self.normalize:
            frame = normalize_pixels_range(frame)

        return frame


def video_infos_ffmpeg(filepath):
    r"""
    Parses the information from a video using ffmpeg.
    Uses subprocess to get the information through a pipe.

    Parameters
    ----------
    filepath : `Path`
        absolute path to the video file which information to extract

    Returns
    -------
    infos : `dict`
        keys are width, height (size of the frames)
        duration (duration of the video in seconds)
        n_frames
    """
    warnings.warn('Estimating number of frames using ffmpeg duration which '
                  'may be inaccurate for certain types of encodings. Try '
                  'setting the MENPO_FFPROBE_CMD environment variable to '
                  'define the path to ffprobe.')

    # Read information using ffmpeg - the call below intentionally causes
    # an error about no output from FFMPEG in order to terminate faster - hence
    # reading the output from stderr.
    command = [_FFMPEG_CMD(), '-i', str(filepath), '-']
    with _call_subprocess(sp.Popen(command, stdout=DEVNULL,
                                   stderr=sp.PIPE)) as pipe:
        raw_infos = pipe.stderr.read().decode()

    # Note: we use '\d+\.?\d*' so we can match both int and float for the fps
    video_info = re.search(
        r"Video:.*(?P<width> \d+)x(?P<height>\d+).*(?P<fps> \d+\.?\d*) fps",
        raw_infos, re.DOTALL).groupdict()

    # Some videos don't have a valid duration
    time = re.search(
        r"Duration:\s(?P<hours>\d+?):(?P<minutes>\d+?):(?P<seconds>\d+\.\d+?),",
        raw_infos, re.DOTALL)
    if time is None:
        raise ValueError('Unable to determine duration for video - please '
                         'install and use ffprobe for accurate frame count '
                         'computation.')

    # Get the duration in seconds and convert size to ints
    time = time.groupdict()
    hours = float(time['hours'])
    minutes = float(time['minutes'])
    seconds = float(time['seconds'])
    duration = 60*60*hours + 60*minutes + seconds

    fps = round(float(video_info['fps']))
    n_frames = round(duration*fps)
    width = int(video_info['width'])
    height = int(video_info['height'])

    # Create the resulting dictionary
    infos = {'duration': duration, 'width': width, 'height': height,
             'n_frames': n_frames, 'fps': fps}

    return infos


def video_infos_ffprobe(filepath):
    """
    Parses the information from a video using ffprobe
    Uses subprocess to get the information through a pipe

    Parameters
    ----------
    filepath : `Path`
        Absolute path to the video file which information to extract

    Returns
    -------
    infos : `dict`
        keys are width, height (size of the frames)
        duration (duration of the video in seconds)
        n_frames
    """
    expected_keys = {'width', 'height', 'avg_frame_rate', 'duration',
                     'nb_read_frames'}

    p = sp.Popen(
        [_FFPROBE_CMD(), '-v', 'quiet',        # Quiet output
         '-count_frames',                      # Count the number of complete frames
         '-select_streams', 'v:0',             # Only show the first stream
         '-show_entries',                      # Show the entries below
         'stream=height,width,nb_read_frames,duration,avg_frame_rate',
         '-of', 'default=noprint_wrappers=1',  # Output format is just to print key=value pairs
         str(filepath)],
        stdin=sp.PIPE,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
    )
    with _call_subprocess(p) as pipe:
        stdout_output = pipe.stdout.readlines()
    del p

    # Split the key=value pairs on the '='
    kv_dict = dict([v.decode().strip().split('=') for v in stdout_output])
    found_keys = set(kv_dict.keys())

    if found_keys ^ expected_keys:
        raise ValueError('Not all of the expected values were returned. '
                         'Expected {} but found {}.'.format(
            expected_keys, found_keys))

    kv_dict['n_frames'] = int(kv_dict.pop('nb_read_frames'))
    kv_dict['width'] = int(kv_dict['width'])
    kv_dict['height'] = int(kv_dict['height'])
    # Some videos may not have a valid duration
    if kv_dict['duration'] == 'N/A':
        kv_dict['duration'] = None
    else:
        kv_dict['duration'] = float(kv_dict['duration'])
    fps = kv_dict.pop('avg_frame_rate').split('/')
    try:
        kv_dict['fps'] = float(fps[0]) / float(fps[1])
    except ZeroDivisionError:
        kv_dict['fps'] = None

    return kv_dict



import numpy as np
from copy import deepcopy
from menpo.base import Copyable
from menpo.transform.base import Alignment, Invertible, Transform
# TODO View is broken for PWA (TriangleContainmentError)


class TriangleContainmentError(Exception):
    r"""
    Exception that is thrown when an attempt is made to map a point with a
    PWATransform that does not lie in a source triangle.

    points_outside_source_domain : ``(d,)`` `ndarray`
        A `bool` value for the ``d`` points that were attempted to be applied.
        If ``True```, the point was outside of the domain.
    """
    def __init__(self, points_outside_source_domain):
        super(TriangleContainmentError, self).__init__()
        self.points_outside_source_domain = points_outside_source_domain


def containment_from_alpha_beta(alpha, beta):
    r"""
    Check `alpha` and `beta` are within a triangle (``alpha >= 0``,
    ``beta >= 0``, ``alpha + beta <= 1``). Returns the indices of the triangles
    that are `alpha` and `beta` are in. If any of the points are not contained
    in a triangle, raises a `TriangleContainmentError`.

    Parameters
    ----------
    alpha : ``(K, n_tris)`` `ndarray`
        Alpha for each point and triangle being tested.
    beta : ``(K, n_tris)`` `ndarray`
        Beta for each point and triangle being tested.

    Returns
    -------
    tri_index : ``(L,)`` `ndarray`
        Triangle index for each `points`, assigning each
        point in a triangle to the triangle index.

    Raises
    ------
    TriangleContainmentError
        All `points` must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
    """
    # (K, n_tris), boolean for whether a given triangle contains a given
    #  point
    point_containment = np.logical_and(np.logical_and(
        alpha >= 0, beta >= 0),
        alpha + beta <= 1)
    # is each point in a triangle?
    point_in_a_triangle = np.any(point_containment, axis=1)
    if np.any(~point_in_a_triangle):
        raise TriangleContainmentError(~point_in_a_triangle)
    point_index, tri_index = np.nonzero(point_containment)
    # don't want duplicates! ensure that here:
    index = np.zeros(alpha.shape[0])
    index[point_index] = tri_index
    return index.astype(np.uint32)


def alpha_beta(i, ij, ik, points):
    r"""
    Calculates the `alpha` and `beta` values (barycentric coordinates) for each
    triangle for all points provided. Note that this does not raise a
    `TriangleContainmentError`.

    Parameters
    ----------
    i : ``(n_tris, 2)`` `ndarray`
        The coordinate of the i'th point of each triangle
    ij : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the j'th point of each
        triangle
    ik : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the k'th point of each
        triangle
    points : ``(n_points, 2)`` `ndarray`
        Points to calculate the barycentric coordinates for.

    Returns
    -------
    alpha : ``(n_points, n_tris)`` `ndarray`
        The `alpha` for each point and triangle. Alpha can be interpreted
        as the contribution of the `ij` vector to the position of the point in
        question.
    beta : ``(n_points, n_tris)`` `ndarray`
        The beta for each point and triangle. Beta can be interpreted as
        the contribution of the ik vector to the position of the point in
        question.
    """
    ip = points[..., None] - i
    dot_jj = np.einsum('dt, dt -> t', ij, ij)
    dot_kk = np.einsum('dt, dt -> t', ik, ik)
    dot_jk = np.einsum('dt, dt -> t', ij, ik)
    dot_pj = np.einsum('vdt, dt -> vt', ip, ij)
    dot_pk = np.einsum('vdt, dt -> vt', ip, ik)

    d = 1.0/(dot_jj * dot_kk - dot_jk * dot_jk)
    alpha = (dot_kk * dot_pj - dot_jk * dot_pk) * d
    beta = (dot_jj * dot_pk - dot_jk * dot_pj) * d
    return alpha, beta


def index_alpha_beta(i, ij, ik, points):
    """
    Finds for each input point the index of it's bounding triangle and the
    `alpha` and `beta` value for that point in the triangle. Note this means
    that the following statements will always be true::

        alpha + beta <= 1
        alpha >= 0
        beta >= 0

    for each triangle result.

    Trying to map a point that does not exist in a triangle throws a
    `TriangleContainmentError`.

    Parameters
    ----------
    i : ``(n_tris, 2)`` `ndarray`
        The coordinate of the i'th point of each triangle
    ij : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the j'th point of each
        triangle
    ik : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the k'th point of each
        triangle
    points : ``(n_points, 2)`` `ndarray`
        Points to calculate the barycentric coordinates for.

    Returns
    -------
    tri_index : ``(n_tris,)`` `ndarray`
        Triangle index for each of the `points`, assigning each point to its
        containing triangle.
    alpha : ``(n_tris,)`` `ndarray`
        Alpha for containing triangle of each point.
    beta : ``(n_tris,)`` `ndarray`
        Beta for containing triangle of each point.

    Raises
    ------
    TriangleContainmentError
        All `points` must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
    """
    alpha, beta = alpha_beta(i, ij, ik, points)
    each_point = np.arange(points.shape[0])
    index = containment_from_alpha_beta(alpha, beta)
    return index, alpha[each_point, index], beta[each_point, index]


def barycentric_vectors(points, trilist):
    r"""
    Compute the affine transformation between each triangle in the `source`
    and `target`. This is calculated analytically.

    Parameters
    ----------
    points : ``(n_points, 2)`` `ndarray`
        Points to calculate the barycentric coordinates for.
    trilist: ``(n_tris, 3)`` `ndarray`
        The 0-based index triangulation joining the points.

    Returns
    -------
    i : ``(n_tris, 2)`` `ndarray`
        The coordinate of the i'th point of each triangle
    ij : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the j'th point of each
        triangle
    ik : ``(n_tris, 2)`` `ndarray`
        The vector between the i'th point and the k'th point of each
        triangle
    """
    # we permute the axes of the indexed point set to have shape
    # [3, n_dims, n_tris] for ease of indexing in.
    x = np.transpose(points[trilist],  axes=[1, 2, 0])
    return x[0], x[1] - x[0], x[2] - x[0]


# Note we inherit from Alignment first to get it's n_dims behavior
class AbstractPWA(Alignment, Transform, Invertible):
    r"""
    A piecewise affine transformation.

    This is composed of a number of triangles defined be a set of `source` and
    `target` vertices. These vertices are related by a common triangle `list`.
    No limitations on the nature of the triangle `list` are imposed. Points can
    then be mapped via barycentric coordinates from the `source` to the `target`
    space. Trying to map points that are not contained by any source triangle
    throws a `TriangleContainmentError`, which contains diagnostic information.

    Parameters
    ----------
    source : :map:`PointCloud` or :map:`TriMesh`
        The source points. If a TriMesh is provided, the triangulation on
        the TriMesh is used. If a PointCloud is provided, a Delaunay
        triangulation of the source is performed automatically.
    target : :map:`PointCloud`
        The target points. Note that the trilist is entirely decided by the
        source.

    Raises
    ------
    ValueError
        Source and target must both be 2D.
    TriangleContainmentError
        All points to apply must be contained in a source triangle. Check
        `error.points_outside_source_domain` to handle this case.
    """
    def __init__(self, source, target):
        from menpo.shape import TriMesh  # to avoid circular import
        if not isinstance(source, TriMesh):
            source = TriMesh(source.points)
        Alignment.__init__(self, source, target)
        if self.n_dims != 2:
            raise ValueError("source and target must be 2 "
                             "dimensional")
        self.ti, self.tij, self.tik = None, None, None
        self._rebuild_target_vectors()

    @property
    def n_tris(self):
        r"""
        The number of triangles in the triangle list.

        :type: `int`
        """
        return self.source.n_tris

    @property
    def trilist(self):
        r"""
        The triangle list.

        :type: ``(n_tris, 3)`` `ndarray`
        """
        return self.source.trilist

    def _rebuild_target_vectors(self):
        r"""
        Rebuild the vectors that are used in the apply method. This needs to
        be called whenever the target is changed.
        """
        t = self.target.points[self.trilist]
        # get vectors ij ik for the target
        self.tij, self.tik = t[:, 1] - t[:, 0], t[:, 2] - t[:, 0]
        # target i'th vertex positions
        self.ti = t[:, 0]

    def _sync_state_from_target(self):
        r"""
        PWA is particularly efficient to sync from target - we don't have to
        do much at all, just rebuild the target vectors.
        """
        self._rebuild_target_vectors()

    def _apply(self, x, **kwargs):
        """
        Applies this transform to a new set of vectors.

        Parameters
        ----------
        x : ``(K, 2)`` `ndarray`
            Points to apply this transform to.

        Returns
        -------
        transformed : ``(K, 2)`` `ndarray`
            The transformed array.
        """
        tri_index, alpha, beta = self.index_alpha_beta(x)
        return (self.ti[tri_index] +
                alpha[:, None] * self.tij[tri_index] +
                beta[:, None] * self.tik[tri_index])

    def _apply_batched(self, x, batch_size, **kwargs):
        # This is a rare case where we need to override the batched apply
        # method. In this case, we override it because we want to the
        # possibly raised TriangleContainmentError to contain ALL the points
        # that were considered, and not just the first batch of points.
        if batch_size is None:
            return self._apply(x, **kwargs)
        else:
            outputs = []
            points_outside_source_domain = []
            n_points = x.shape[0]
            exception_thrown = False
            for lo_ind in range(0, n_points, batch_size):
                try:
                    hi_ind = lo_ind + batch_size
                    outputs.append(self._apply(x[lo_ind:hi_ind], **kwargs))
                except TriangleContainmentError as e:
                    exception_thrown = True
                    points_outside_source_domain.append(
                        e.points_outside_source_domain)
                else:
                    # No exception was thrown, so all points were inside
                    points_outside_source_domain.append(
                        np.zeros(batch_size, dtype=np.bool))

            if exception_thrown:
                raise TriangleContainmentError(
                    np.hstack(points_outside_source_domain))
            else:
                return np.vstack(outputs)

    def index_alpha_beta(self, points):
        """
        Finds for each input point the index of its bounding triangle and the
        `alpha` and `beta` value for that point in the triangle. Note this
        means that the following statements will always be true::

            alpha + beta <= 1
            alpha >= 0
            beta >= 0

        for each triangle result.

        Trying to map a point that does not exist in a triangle throws a
        `TriangleContainmentError`.

        Parameters
        ----------
        points : ``(K, 2)`` `ndarray`
            Points to test.

        Returns
        -------
        tri_index : ``(L,)`` `ndarray`
            Triangle index for each of the `points`, assigning each
            point to it's containing triangle.
        alpha : ``(L,)`` `ndarray`
            Alpha for containing triangle of each point.
        beta : ``(L,)`` `ndarray`
            Beta for containing triangle of each point.

        Raises
        ------
        TriangleContainmentError
            All `points` must be contained in a source triangle. Check
            `error.points_outside_source_domain` to handle this case.
        """
        raise NotImplementedError()

    @property
    def has_true_inverse(self):
        """
        The inverse is true.

        :type: ``True``
        """
        return True

    def pseudoinverse(self):
        r"""
        The pseudoinverse of the transform - that is, the transform that
        results from swapping `source` and `target`, or more formally, negating
        the transforms parameters. If the transform has a true inverse this
        is returned instead.

        :type: ``type(self)``
        """
        from menpo.shape import PointCloud, TriMesh  # to avoid circular import
        new_source = TriMesh(self.target.points, self.source.trilist)
        new_target = PointCloud(self.source.points)
        return type(self)(new_source, new_target)


class PythonPWA(AbstractPWA):

    def __init__(self, source, target):
        super(PythonPWA, self).__init__(source, target)
        si, sij, sik = barycentric_vectors(self.source.points, self.trilist)
        self.s, self.sij, self.sik = si, sij, sik

    def index_alpha_beta(self, points):
        return index_alpha_beta(self.s, self.sij, self.sik, points)


class CachedPWA(PythonPWA):

    def __init__(self, source, target):
        super(CachedPWA, self).__init__(source, target)
        self._applied_points, self._iab = None, None

    def index_alpha_beta(self, points):
        if (self._applied_points is None or
                not points.shape == self._applied_points.shape or
                not np.allclose(points, self._applied_points)):
            # This must happen first in case index_alpha_beta throws a
            # TriangleContainmentError
            self._iab = PythonPWA.index_alpha_beta(self, points)
            self._applied_points = points
        return self._iab



import numpy as np
cimport numpy as np
cimport cython
from ..cy_utils cimport dtype_from_memoryview


ctypedef fused IMAGE_TYPES:
    float
    double
    np.uint8_t
    np.uint16_t


ctypedef fused CENTRE_TYPES:
    float
    double


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void calc_augmented_centers(CENTRE_TYPES[:, :] centres,
                                 Py_ssize_t[:, :] offsets,
                                 Py_ssize_t[:, :] augmented_centers):
    cdef Py_ssize_t total_index = 0, i = 0, j = 0

    for i in range(centres.shape[0]):
        for j in range(offsets.shape[0]):
            augmented_centers[total_index, 0] = <Py_ssize_t> (centres[i, 0] + offsets[j, 0])
            augmented_centers[total_index, 1] = <Py_ssize_t> (centres[i, 1] + offsets[j, 1])
            total_index += 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void calc_slices(Py_ssize_t[:, :] centres,
                      Py_ssize_t image_shape0,
                      Py_ssize_t image_shape1,
                      Py_ssize_t patch_shape0,
                      Py_ssize_t patch_shape1,
                      Py_ssize_t half_patch_shape0,
                      Py_ssize_t half_patch_shape1,
                      Py_ssize_t add_to_patch0,
                      Py_ssize_t add_to_patch1,
                      Py_ssize_t[:, :] ext_s_min,
                      Py_ssize_t[:, :] ext_s_max,
                      Py_ssize_t[:, :] ins_s_min,
                      Py_ssize_t[:, :] ins_s_max):
    cdef Py_ssize_t i = 0

    for i in range(centres.shape[0]):
        c_min_new0 = centres[i, 0] - half_patch_shape0
        c_min_new1 = centres[i, 1] - half_patch_shape1
        c_max_new0 = centres[i, 0] + half_patch_shape0 + add_to_patch0
        c_max_new1 = centres[i, 1] + half_patch_shape1 + add_to_patch1

        ext_s_min[i, 0] = c_min_new0
        ext_s_min[i, 1] = c_min_new1
        ext_s_max[i, 0] = c_max_new0
        ext_s_max[i, 1] = c_max_new1

        if ext_s_min[i, 0] < 0:
            ext_s_min[i, 0] = 0
        if ext_s_min[i, 1] < 0:
            ext_s_min[i, 1] = 0
        if ext_s_min[i, 0] > image_shape0:
            ext_s_min[i, 0] = image_shape0 - 1
        if ext_s_min[i, 1] > image_shape1:
            ext_s_min[i, 1] = image_shape1 - 1

        if ext_s_max[i, 0] < 0:
            ext_s_max[i, 0] = 0
        if ext_s_max[i, 1] < 0:
            ext_s_max[i, 1] = 0
        if ext_s_max[i, 0] > image_shape0:
            ext_s_max[i, 0] = image_shape0 - 1
        if ext_s_max[i, 1] > image_shape1:
            ext_s_max[i, 1] = image_shape1 - 1

        ins_s_min[i, 0] = ext_s_min[i, 0] - c_min_new0
        ins_s_min[i, 1] = ext_s_min[i, 1] - c_min_new1

        ins_s_max[i, 0] = ext_s_max[i, 0] - c_max_new0 + patch_shape0
        if ins_s_max[i, 0] < 0:
            ins_s_max[i, 0] = 0
        ins_s_max[i, 1] = ext_s_max[i, 1] - c_max_new1 + patch_shape1
        if ins_s_max[i, 1] < 0:
            ins_s_max[i, 1] = 0


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef extract_patches(IMAGE_TYPES[:, :, :] image,
                      CENTRE_TYPES[:, :] centres,
                      Py_ssize_t[:] patch_shape, Py_ssize_t[:, :] offsets):
    dtype = dtype_from_memoryview(image)
    cdef:
        Py_ssize_t n_centres = centres.shape[0]
        Py_ssize_t n_offsets = offsets.shape[0]
        Py_ssize_t n_augmented_centres = n_centres * n_offsets
        object extents_size = [n_augmented_centres, 2]

        Py_ssize_t half_patch_shape0 = patch_shape[0] // 2
        Py_ssize_t half_patch_shape1 = patch_shape[1] // 2
        Py_ssize_t add_to_patch0 = patch_shape[0] % 2
        Py_ssize_t add_to_patch1 = patch_shape[1] % 2
        Py_ssize_t patch_shape0 = patch_shape[0]
        Py_ssize_t patch_shape1 = patch_shape[1]
        Py_ssize_t image_shape0 = image.shape[1]
        Py_ssize_t image_shape1 = image.shape[2]
        Py_ssize_t n_channels = image.shape[0]

        Py_ssize_t total_index = 0, i = 0, j = 0

        # Although it is faster to use malloc in this case, the change in syntax
        # and the mental overhead of handling freeing memory is not considered
        # worth it for these buffers. From simple tests it seems you only begin
        # to see a performance difference when you have
        # n_augmented_centres >~ 5000
        Py_ssize_t[:, :] augmented_centers = np.empty(extents_size,
                                                      dtype=np.intp)
        Py_ssize_t[:, :] ext_s_max = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ext_s_min = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ins_s_max = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ins_s_min = np.empty(extents_size, dtype=np.intp)

        np.ndarray[IMAGE_TYPES, ndim=5] patches = np.zeros(
            [n_centres, n_offsets, n_channels, patch_shape0, patch_shape1],
            dtype=dtype)

    calc_augmented_centers(centres, offsets, augmented_centers)
    calc_slices(augmented_centers, image_shape0, image_shape1, patch_shape0,
                patch_shape1, half_patch_shape0, half_patch_shape1,
                add_to_patch0, add_to_patch1, ext_s_min, ext_s_max, ins_s_min,
                ins_s_max)

    for i in range(n_centres):
        for j in range(n_offsets):
            patches[i, j, :,
                ins_s_min[total_index, 0]:ins_s_max[total_index, 0],
                ins_s_min[total_index, 1]:ins_s_max[total_index, 1]
            ] = \
                image[:,
                    ext_s_min[total_index, 0]:ext_s_max[total_index, 0],
                    ext_s_min[total_index, 1]:ext_s_max[total_index, 1]]
            total_index += 1

    return patches


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void set_patches(IMAGE_TYPES[:, :, :, :, :] patches,
                       IMAGE_TYPES[:, :, :] image,
                       CENTRE_TYPES[:, :] centres,
                       Py_ssize_t[:, :] offset,
                       Py_ssize_t offset_index):
    cdef:
        Py_ssize_t n_centres = centres.shape[0]
        object extents_size = [n_centres, 2]

        Py_ssize_t patch_shape0 = patches.shape[3]
        Py_ssize_t patch_shape1 = patches.shape[4]
        Py_ssize_t half_patch_shape0 = patch_shape0 // 2
        Py_ssize_t half_patch_shape1 = patch_shape1 // 2
        Py_ssize_t add_to_patch0 = patch_shape0 % 2
        Py_ssize_t add_to_patch1 = patch_shape1 % 2
        Py_ssize_t image_shape0 = image.shape[1]
        Py_ssize_t image_shape1 = image.shape[2]
        Py_ssize_t n_channels = image.shape[0]

        Py_ssize_t total_index = 0, i = 0

        # Although it is faster to use malloc in this case, the change in syntax
        # and the mental overhead of handling freeing memory is not considered
        # worth it for these buffers. From simple tests it seems you only begin
        # to see a performance difference when you have
        # n_augmented_centres >~ 5000
        Py_ssize_t[:, :] augmented_centers = np.empty(extents_size,
                                                      dtype=np.intp)
        Py_ssize_t[:, :] ext_s_max = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ext_s_min = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ins_s_max = np.empty(extents_size, dtype=np.intp)
        Py_ssize_t[:, :] ins_s_min = np.empty(extents_size, dtype=np.intp)

    calc_augmented_centers(centres, offset, augmented_centers)
    calc_slices(augmented_centers, image_shape0, image_shape1, patch_shape0,
                patch_shape1, half_patch_shape0, half_patch_shape1,
                add_to_patch0, add_to_patch1, ext_s_min, ext_s_max, ins_s_min,
                ins_s_max)

    for i in range(n_centres):
        image[:,
              ext_s_min[total_index, 0]:ext_s_max[total_index, 0],
              ext_s_min[total_index, 1]:ext_s_max[total_index, 1]
             ] = \
                 patches[i, offset_index, :,
                         ins_s_min[total_index, 0]:ins_s_max[total_index, 0],
                         ins_s_min[total_index, 1]:ins_s_max[total_index, 1]]
        total_index += 1



import numpy as np

from .base import HomogFamilyAlignment
from .affine import DiscreteAffine
from .similarity import Similarity


class Translation(DiscreteAffine, Similarity):
    r"""
    An ``n_dims``-dimensional translation transform.

    Parameters
    ----------
    translation : ``(n_dims,)`` `ndarray`
        The translation in each axis.
    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``h_matrix`` for performance.
    """

    def __init__(self, translation, skip_checks=False):
        translation = np.asarray(translation)
        h_matrix = np.eye(translation.shape[0] + 1)
        h_matrix[:-1, -1] = translation
        Similarity.__init__(self, h_matrix, copy=False,
                            skip_checks=skip_checks)

    @classmethod
    def init_identity(cls, n_dims):
        r"""
        Creates an identity transform.

        Parameters
        ----------
        n_dims : `int`
            The number of dimensions.

        Returns
        -------
        identity : :class:`Translation`
            The identity matrix transform.
        """
        return Translation(np.zeros(n_dims))

    def _transform_str(self):
        message = 'Translation by {}'.format(self.translation_component)
        return message

    @property
    def n_parameters(self):
        r"""
        The number of parameters: ``n_dims``

        :type: `int`
        """
        return self.n_dims

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. The parameters
        are output in the order ``[t0, t1, ...]``.

        +-----------+--------------------------------------------+
        |parameter | definition                                  |
        +==========+=============================================+
        |t0        | The translation in the first axis           |
        |t1        | The translation in the second axis          |
        |...       | ...                                         |
        |tn        | The translation in the nth axis             |
        +----------+---------------------------------------------+

        Returns
        -------
        ts : ``(n_dims,)`` `ndarray`
            The translation in each axis.
        """
        return self.h_matrix[:-1, -1]

    def _from_vector_inplace(self, p):
        r"""
        Updates the :class:`Translation` inplace.

        Parameters
        ----------
        vector : ``(n_dims,)`` `ndarray`
            The array of parameters.
        """
        self.h_matrix[:-1, -1] = p

    def pseudoinverse(self):
        r"""
        The inverse translation (negated).

        :type: :class:`Translation`
        """
        return Translation(-self.translation_component, skip_checks=True)


class AlignmentTranslation(HomogFamilyAlignment, Translation):
    r"""
    Constructs a :class:`Translation` by finding the optimal translation
    transform to align `source` to `target`.

    Parameters
    ----------
    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment
    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment
    """

    def __init__(self, source, target):
        HomogFamilyAlignment.__init__(self, source, target)
        Translation.__init__(self, target.centre() - source.centre())

    def _from_vector_inplace(self, p):
        r"""
        Updates the :class:`Translation` inplace.

        Parameters
        ----------
        vector : ``(n_dims,)`` `ndarray`
            The array of parameters.
        """
        Translation._from_vector_inplace(self, p)
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        translation = self.target.centre() - self.source.centre()
        self.h_matrix[:-1, -1] = translation

    def as_non_alignment(self):
        r"""
        Returns a copy of this translation without its alignment nature.

        Returns
        -------
        transform : :map:`Translation`
            A version of this transform with the same transform behavior but
            without the alignment logic.
        """
        return Translation(self.translation_component)



import numpy as np
from pytest import raises
from numpy.testing import assert_allclose

from menpo.shape import PointCloud
from menpo.image import MaskedImage, BooleanImage


def test_init_from_pointcloud_constrain_mask():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = MaskedImage.init_from_pointcloud(pc, constrain_mask=True)
    assert im.shape == (15, 15)
    assert im.mask.n_true() == 120


def test_init_from_pointcloud_no_constrain_mask():
    pc = PointCloud(np.array([[5, 5], [5, 20], [20, 20]]))
    im = MaskedImage.init_from_pointcloud(pc, constrain_mask=False)
    assert im.shape == (15, 15)
    assert im.mask.n_true() == 225


def test_constrain_mask_to_landmarks_pwa():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0.0, 0.0], [5.0, 0.0],
                                                [5.0, 5.0], [0.0, 5.0]]))
    new_img = img.constrain_mask_to_landmarks(group='box')

    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert(new_img.mask.n_true() == 36)
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_pwa_batched():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0.0, 0.0], [5.0, 0.0],
                                                [5.0, 5.0], [0.0, 5.0]]))
    new_img = img.constrain_mask_to_landmarks(group='box', batch_size=2)

    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert(new_img.mask.n_true() == 36)
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_convex_hull():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0.], [5., 0.],
                                                [5., 5.], [0., 5.]]))
    new_img = img.constrain_mask_to_landmarks(group='box',
                                              point_in_pointcloud='convex_hull')
    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, 1:6] = True
    assert(new_img.mask.n_true() == 30)
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_callable():
    def bounding_box(_, indices):
        return np.ones(indices.shape[0], dtype=np.bool)

    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0.], [5., 0.],
                                                [5., 5.], [0., 5.]]))
    new_img = img.constrain_mask_to_landmarks(group='box',
                                              point_in_pointcloud=bounding_box)
    example_mask = BooleanImage.init_blank((10, 10), fill=False)
    example_mask.pixels[0, :6, :6] = True
    assert(new_img.mask.n_true() == 36)
    assert_allclose(new_img.mask.pixels, example_mask.pixels)


def test_constrain_mask_to_landmarks_non_2d():
    img = MaskedImage.init_blank((10, 10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0., 0.]]))
    with raises(ValueError):
        img.constrain_mask_to_landmarks()


def test_constrain_mask_to_landmarks_unknown_key():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0., 0.]]))
    with raises(ValueError):
        img.constrain_mask_to_landmarks(point_in_pointcloud='unknown')


def test_erode():
    img = MaskedImage.init_blank((10, 10))
    img2 = img.erode()
    assert(img2.mask.n_true() == 64)
    img3 = img.erode(n_pixels=3)
    assert(img3.mask.n_true() == 16)


def test_constrain_mask_to_patches_around_landmarks_even():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0.], [5., 0.],
                                                [5., 5.], [0., 5.]]))
    new_img = img.constrain_mask_to_patches_around_landmarks((2,2), group='box')
    assert(new_img.mask.n_true() == 9)
    assert_allclose(new_img.mask.pixels[:, 0, 0], True)
    assert_allclose(new_img.mask.pixels[:, 4:6, 0], True)
    assert_allclose(new_img.mask.pixels[:, 0, 4:6], True)
    assert_allclose(new_img.mask.pixels[:, 4:6, 4:6], True)


def test_constrain_mask_to_patches_around_landmarks_odd():
    img = MaskedImage.init_blank((10, 10))
    img.landmarks['box'] = PointCloud(np.array([[0., 0.], [5., 0.],
                                                [5., 5.], [0., 5.]]))
    new_img = img.constrain_mask_to_patches_around_landmarks((3,3), group='box')
    assert(new_img.mask.n_true() == 25)
    assert_allclose(new_img.mask.pixels[:, :2, :2], True)
    assert_allclose(new_img.mask.pixels[:, 4:7, :2], True)
    assert_allclose(new_img.mask.pixels[:, :2, 4:7], True)
    assert_allclose(new_img.mask.pixels[:, 4:7, 4:7], True)


def test_set_boundary_pixels():
    mask = np.ones((10, 10), dtype=np.bool)
    img = MaskedImage.init_blank((10, 10), mask=mask, fill=0., n_channels=1)
    new_img = img.set_boundary_pixels(value=2.)
    assert(new_img.mask.n_true() == 100)
    assert(~np.allclose(img.pixels, new_img.pixels))
    assert_allclose(new_img.pixels[0, 1:-1, 1:-1], 0.)
    assert_allclose(new_img.pixels[0, :, 0],       2.)
    assert_allclose(new_img.pixels[0, 0, :],       2.)
    assert_allclose(new_img.pixels[0, :, -1],      2.)
    assert_allclose(new_img.pixels[0, -1, :],      2.)


def test_dilate():
    img = MaskedImage.init_blank((10, 10))
    img = img.erode(n_pixels=3)
    img2 = img.dilate()
    assert(img2.mask.n_true() == 32)
    img3 = img.dilate(n_pixels=3)
    assert(img3.mask.n_true() == 76)


def test_init_from_rolled_channels():
    p = np.empty([50, 60, 3])
    im = MaskedImage.init_from_channels_at_back(p)
    assert im.n_channels == 3
    assert im.height == 50
    assert im.width == 60


def test_init_from_rolled_channels_masked():
    p = np.empty([50, 60, 3])
    example_mask = BooleanImage.init_blank((50, 60), fill=False)
    example_mask.pixels[0, :6, :6] = True

    im = MaskedImage.init_from_channels_at_back(p, mask=example_mask)
    assert im.n_channels == 3
    assert im.height == 50
    assert im.width == 60
    assert im.mask.n_true() == 36
